import json
from typing import Dict, Any

from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import utils
from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand

__all__ = ['TransferAssetCommand']


class TransferAssetCommand(NeoExpressCommand):
    def __init__(self, token_symbol: str, sender: Account, receiver: Account, quantity: float,
                 data: Any = None,
                 trace: bool = False,
                 decimals: int = 0,
                 neo_express_data_file: str = None):

        self.token_symbol = token_symbol
        self.sender = sender
        self.receiver = receiver
        self.quantity = quantity

        self.data = data
        self.trace = trace
        self.input = neo_express_data_file

        if decimals < 0:
            decimals = 0

        command_id = 'transfer'
        args = [
            utils.stringify_asset_quantity(quantity, decimals),
            token_symbol,
            sender.get_identifier(),
            receiver.get_identifier(),
        ]

        super().__init__(command_id, args)

    def _get_options(self) -> Dict[str, str]:
        options = super()._get_options()

        if self.data is not None:
            options['--data'] = json.dumps(self.data)
        if self.trace:
            options['--trace'] = ''

        return options



from enum import IntFlag


class FindOptions(IntFlag):
    """
    Represents the options you can use when trying to find a set of values inside the storage.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/services/FindOptions>`__
    to learn more about the FindOption class.
    """
    NONE = 0
    """
    No option is set. The results will be an iterator of (key, value).

    :meta hide-value:
    """

    KEYS_ONLY = 1 << 0
    """
    Indicates that only keys need to be returned. The results will be an iterator of keys.

    :meta hide-value:
    """

    REMOVE_PREFIX = 1 << 1
    """
    Indicates that the prefix byte of keys should be removed before return.

    :meta hide-value:
    """

    VALUES_ONLY = 1 << 2
    """
    Indicates that only values need to be returned. The results will be an iterator of values.

    :meta hide-value:
    """

    DESERIALIZE_VALUES = 1 << 3
    """
    Indicates that values should be deserialized before return.

    :meta hide-value:
    """

    PICK_FIELD_0 = 1 << 4
    """
    Indicates that only the field 0 of the deserialized values need to be returned.
    This flag must be set together with DESERIALIZE_VALUES, e.g., `DESERIALIZE_VALUES | PICK_FIELD_0`

    :meta hide-value:
    """

    PICK_FIELD_1 = 1 << 5
    """
    Indicates that only the field 1 of the deserialized values need to be returned.
    This flag must be set together with DESERIALIZE_VALUES, e.g., `DESERIALIZE_VALUES | PICK_FIELD_1`

    :meta hide-value:
    """



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.internal import constants
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class UInt160Method(IBuiltinMethod):

    def __init__(self, return_type: UInt160Type, argument_type: IType = None):
        from boa3.internal.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.none

        identifier = 'UInt160'
        args: Dict[str, Variable] = {'object': Variable(argument_type)}

        args_default = ast.parse("{0}".format(Type.int.default_value)
                                 ).body[0].value

        super().__init__(identifier, args, [args_default], return_type=return_type)

    @property
    def _arg_object(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_object.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_object.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.internal.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.internal.model.type.type import Type

        return (Type.bytes.is_type_of(param_type)
                or Type.int.is_type_of(param_type))

    def evaluate_literal(self, *args: Any) -> Any:
        from boa3.internal.neo3.core.types import UInt160

        if len(args) == 0:
            return UInt160.zero().to_array()

        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, int):
                from boa3.internal.neo.vm.type.Integer import Integer
                arg = Integer(arg).to_byte_array(min_length=UInt160._BYTE_LEN)
            if isinstance(arg, bytes):
                value = UInt160(arg).to_array()
                return value

        return super().evaluate_literal(*args)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.neo.vm.type.Integer import Integer
        from boa3.internal.model.type.type import Type

        from boa3.internal.neo.vm.type.StackItem import StackItemType
        return [
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, Type.int.stack_item),  # if istype(arg, int):
            (Opcode.JMPIFNOT, Integer(46).to_byte_array(signed=True)),
            (Opcode.DUP, b''),                       # assert num >= 0
            (Opcode.PUSH0, b''),
            (Opcode.GE, b''),
            (Opcode.ASSERT, b''),

            (Opcode.DUP, b''),                       # if len(num) < 20
            (Opcode.SIZE, b''),                        # increase number's length to 20
            (Opcode.PUSHINT8, Integer(constants.SIZE_OF_INT160).to_byte_array(signed=True)),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),

            (Opcode.JMPGE, Integer(30).to_byte_array(signed=True)),
            (Opcode.PUSHDATA1, (Integer(constants.SIZE_OF_INT160).to_byte_array(signed=True)
                                + bytes(constants.SIZE_OF_INT160))),
            (Opcode.REVERSE3, b''),
            (Opcode.SUB, b''),
            (Opcode.LEFT, b''),
            (Opcode.CAT, b''),

            (Opcode.JMP, Integer(4).to_byte_array()),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),

            (Opcode.CONVERT, StackItemType.ByteString),  # convert to uint160
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSHINT8, Integer(constants.SIZE_OF_INT160).to_byte_array(signed=True)),
            (Opcode.NUMEQUAL, b''),
            (Opcode.ASSERT, b''),
        ]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_object.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return UInt160Method(self.return_type, value)
        return super().build(value)



from typing import Dict, Union

from boa3_test.test_drive.neoxp.command import utils
from boa3_test.test_drive.neoxp.command.neoexpresscommand.show.ishowcommand import IShowCommand


class ShowBlockCommand(IShowCommand):
    def __init__(self, block_hash_or_index: Union[bytes, str, int] = None,
                 neo_express_data_file: str = None):

        if isinstance(block_hash_or_index, int) and block_hash_or_index < 0:
            block_hash_or_index = 0

        if isinstance(block_hash_or_index, bytes):
            block_hash_or_index = utils.to_hex_str(block_hash_or_index)

        arguments = []
        if isinstance(block_hash_or_index, (str, int)):
            arguments.append(str(block_hash_or_index))
            hash_or_index = block_hash_or_index
        else:
            hash_or_index = None

        self.block_hash_or_index = hash_or_index
        self.input = neo_express_data_file

        super().__init__('block', arguments)

    def _get_options(self) -> Dict[str, str]:
        return super()._get_options()



from argparse import ArgumentParser, RawDescriptionHelpFormatter

from boa3.internal import constants
from boa3.internal.cli_commands import commands


def main():
    n3boa = ArgumentParser(description=f"neo3-boa by COZ - version {constants.BOA_VERSION}"
                                       "\nWrite smart contracts for Neo3 in Python",
                           formatter_class=RawDescriptionHelpFormatter,
                           )
    n3boa.add_argument("-v", "--version", action="version",
                       version=f"neo3-boa {constants.BOA_VERSION}")

    n3_subparser = n3boa.add_subparsers(title='Commands')

    # Initialize all commands on cli_commands
    for command in commands:
        command(n3_subparser).add_arguments_and_callback()

    # read command line and get the correct command requested
    args = n3boa.parse_args()

    # execute subparser commands
    if hasattr(args, 'func'):
        args.func(vars(args))
    else:
        import sys
        n3boa.print_help(sys.stderr)


if __name__ == "__main__":
    main()



__all__ = ['CheckMultisigMethod',
           'CheckSigMethod',
           'Hash160Method',
           'Hash256Method',
           'Murmur32Method',
           'NamedCurveType',
           'Ripemd160Method',
           'Sha256Method',
           'VerifyWithECDsaMethod',
           ]

from boa3.internal.model.builtin.interop.crypto.checkmultisigmethod import CheckMultisigMethod
from boa3.internal.model.builtin.interop.crypto.checksigmethod import CheckSigMethod
from boa3.internal.model.builtin.interop.crypto.hash160method import Hash160Method
from boa3.internal.model.builtin.interop.crypto.hash256method import Hash256Method
from boa3.internal.model.builtin.interop.crypto.murmur32method import Murmur32Method
from boa3.internal.model.builtin.interop.crypto.namedcurvetype import NamedCurveType
from boa3.internal.model.builtin.interop.crypto.ripemd160method import Ripemd160Method
from boa3.internal.model.builtin.interop.crypto.sha256method import Sha256Method
from boa3.internal.model.builtin.interop.crypto.verifywithecdsa import VerifyWithECDsaMethod



from __future__ import annotations

from boa3_test.test_drive.model.network.payloads.witness import Witness as TestWitness

__all__ = ['Witness']


class Witness(TestWitness):
    def __init__(self, invocation_script: bytes, verification_script: bytes):
        super().__init__(invocation_script, verification_script)



from typing import List, Tuple

from boa3.internal.model.builtin.method.printmethod import PrintMethod
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintBoolMethod(PrintMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        arg_value = Type.bool

        super().__init__(arg_value)

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:
            from boa3.internal.model.builtin.interop.interop import Interop
            self._print_value_opcodes = (
                [
                    (Opcode.NZ, b''),
                ] + Interop.JsonSerialize.opcode
            )

        return super().print_value_opcodes



from typing import Dict

from boa3.internal.model.builtin.builtinproperty import IBuiltinProperty
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.variable import Variable


class GetGasLeftMethod(InteropMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = '-get_gas_left'
        syscall = 'System.Runtime.GasLeft'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=Type.int)


class GasLeftProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'gas_left'
        getter = GetGasLeftMethod()
        super().__init__(identifier, getter)



__all__ = ['AppendMethod',
           'ClearMethod',
           'CopyDictMethod',
           'CopyListMethod',
           'CountSequenceGenericMethod',
           'CountSequencePrimitiveMethod',
           'CountStrMethod',
           'ExtendMethod',
           'IndexBytesStringMethod',
           'IndexSequenceMethod',
           'InsertMethod',
           'IsDigitMethod',
           'JoinMethod',
           'LowerMethod',
           'MapKeysMethod',
           'MapValuesMethod',
           'PopDictMethod',
           'PopDictDefaultMethod',
           'PopSequenceMethod',
           'RemoveMethod',
           'ReverseMethod',
           'StartsWithMethod',
           'StripMethod',
           'UpperMethod',
           'UpperMethod',
           ]

from boa3.internal.model.builtin.classmethod.appendmethod import AppendMethod
from boa3.internal.model.builtin.classmethod.clearmethod import ClearMethod
from boa3.internal.model.builtin.classmethod.copydictmethod import CopyDictMethod
from boa3.internal.model.builtin.classmethod.copylistmethod import CopyListMethod
from boa3.internal.model.builtin.classmethod.countsequencegenericmethod import CountSequenceGenericMethod
from boa3.internal.model.builtin.classmethod.countsequenceprimitivemethod import CountSequencePrimitiveMethod
from boa3.internal.model.builtin.classmethod.countstrmethod import CountStrMethod
from boa3.internal.model.builtin.classmethod.extendmethod import ExtendMethod
from boa3.internal.model.builtin.classmethod.indexbytesstringmethod import IndexBytesStringMethod
from boa3.internal.model.builtin.classmethod.indexsequencemethod import IndexSequenceMethod
from boa3.internal.model.builtin.classmethod.insertmethod import InsertMethod
from boa3.internal.model.builtin.classmethod.isdigitmethod import IsDigitMethod
from boa3.internal.model.builtin.classmethod.joinmethod import JoinMethod
from boa3.internal.model.builtin.classmethod.lowermethod import LowerMethod
from boa3.internal.model.builtin.classmethod.mapkeysmethod import MapKeysMethod
from boa3.internal.model.builtin.classmethod.mapvaluesmethod import MapValuesMethod
from boa3.internal.model.builtin.classmethod.popdictdefaultmethod import PopDictDefaultMethod
from boa3.internal.model.builtin.classmethod.popdictmethod import PopDictMethod
from boa3.internal.model.builtin.classmethod.popsequencemethod import PopSequenceMethod
from boa3.internal.model.builtin.classmethod.removemethod import RemoveMethod
from boa3.internal.model.builtin.classmethod.reversemethod import ReverseMethod
from boa3.internal.model.builtin.classmethod.startswithmethod import StartsWithMethod
from boa3.internal.model.builtin.classmethod.stripmethod import StripMethod
from boa3.internal.model.builtin.classmethod.uppermethod import UpperMethod



from typing import Optional, List, Tuple, Any

from boa3.internal.model.builtin.builtinproperty import IBuiltinProperty
from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class GetEnvMethod(IBuiltinMethod):
    def __init__(self, env: str = None):
        from boa3.internal import constants
        from boa3.internal.model.type.type import Type

        identifier = '-get_env'
        super().__init__(identifier, return_type=Type.str)
        self._env = env if isinstance(env, str) and len(env) > 0 else constants.DEFAULT_CONTRACT_ENVIRONMENT

    @property
    def _args_on_stack(self) -> int:
        return super()._args_on_stack

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.neo.vm.opcode import OpcodeHelper
        return [
            OpcodeHelper.get_pushdata_and_data(self._env)
        ]

    @property
    def _body(self) -> Optional[str]:
        return None

    def update_with_analyser(self, analyser):
        from boa3.internal.analyser.analyser import Analyser
        if isinstance(analyser, Analyser):
            self.reset()
            self._env = analyser.env

        super().update_with_analyser(analyser)

    def build(self, value: Any = None) -> IBuiltinMethod:
        if isinstance(value, str):
            result = GetEnvMethod(value)
        else:
            result = super().build(value)
        return result


class EnvProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'env'
        getter = GetEnvMethod()
        super().__init__(identifier, getter)
        self._getter: GetEnvMethod = self._getter  # ensure typing

    def build(self, env: str = None, *args, **kwargs) -> IBuiltinProperty:
        getter = self._getter.build(env)
        if getter != self._getter:
            result = EnvProperty()
            result._getter = getter
        else:
            result = super().build(env, *args, **kwargs)
        return result


Env = EnvProperty()



from __future__ import annotations

from typing import Any, Dict, List, Optional

from boa3.internal.neo3.vm import VMState, vmstate
from boa3_test.test_drive.model.network.payloads.testtransaction import TestTransaction
from boa3_test.tests.test_classes import transactionattribute as tx_attribute
from boa3_test.tests.test_classes.signer import Signer
from boa3_test.tests.test_classes.witness import Witness


class Transaction(TestTransaction):
    def __init__(self, script: bytes, signers: List[Signer] = None, witnesses: List[Witness] = None):
        super().__init__(None, script, signers, witnesses)
        self._attributes: List[tx_attribute.TransactionAttribute] = []
        self._state: VMState = VMState.NONE

    def add_attribute(self, tx_attr: tx_attribute.TransactionAttribute):
        if tx_attr not in self._attributes:
            self._attributes.append(tx_attr)

    @property
    def hash(self) -> Optional[bytes]:
        if self._hash is None:
            return None

        return super().hash.to_array()

    @property
    def state(self) -> VMState:
        return self._state

    def to_json(self) -> Dict[str, Any]:
        json = super().to_json()

        json.pop('hash')  # don't include hash in this json
        json['state'] = self._state.name

        return json

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Transaction:
        tx: Transaction = super().from_json(json)

        if 'attributes' in json:
            attributes_json = json['attributes']
            if not isinstance(attributes_json, list):
                attributes_json = [attributes_json]
            tx._attributes = [tx_attribute.TransactionAttribute.from_json(js) for js in attributes_json]

        if 'state' in json and isinstance(json['state'], str):
            tx._state = vmstate.get_vm_state(json['state'])

        return tx

    def copy(self):
        copied = Transaction(self._script, self._signers, self._witnesses)
        copied._hash = self._hash
        return copied

    def __eq__(self, other) -> bool:
        if not isinstance(other, Transaction):
            return False
        if self._hash == other._hash:
            return True

        return (self._script == other._script
                and self._attributes == self._attributes
                and self._signers == other._signers
                and self._witnesses == other._witnesses)



from typing import Union

from boa3.internal import constants as __boa_constants
from boa3.internal.neo import to_hex_str

GAS_DECIMALS = __boa_constants.GAS_DECIMALS
NEO_DECIMALS = __boa_constants.NEO_DECIMALS
DEFAULT_ACCOUNT_VERSION = 53


def stringify_asset_quantity(quantity: Union[int, float], decimals: int) -> str:
    if decimals < 0:
        decimals = 0

    quantity_format = f'.{decimals}f'
    formatted_quantity = f'{quantity:{quantity_format}}'  # format to fix decimals
    if decimals > 0:
        formatted_quantity = (
            formatted_quantity.rstrip('0').rstrip('.')  # remove unnecessary zeros in decimals
            .replace('.', __boa_constants.SYS_LOCALE_DECIMAL_POINT)  # fix locale
        )

    return formatted_quantity



from typing import Dict, List, Tuple

from boa3.internal.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class UnVoteMethod(NeoContractMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'un_vote'
        native_identifier = 'vote'  # un_vote is not neo native
        args: Dict[str, Variable] = {
            'account': Variable(UInt160Type.build()),
        }

        neo_internal_args = {
            'account': Variable(UInt160Type.build()),
            'vote_to': Variable(ECPointType.build())
        }

        super().__init__(identifier, native_identifier, args, return_type=Type.bool,
                         internal_call_args=len(neo_internal_args))

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return (
            [
                (Opcode.PUSHNULL, b''),
                (Opcode.SWAP, b''),
            ] +
            super()._opcode
        )



__all__ = [
    'Notification',
    'TriggerType',
    'check_witness',
    'notify',
    'log',
    'get_trigger',
    'get_notifications',
    'get_network',
    'burn_gas',
    'get_random',
    'load_script',
    'address_version',
    'executing_script_hash',
    'calling_script_hash',
    'time',
    'gas_left',
    'platform',
    'invocation_counter',
    'entry_script_hash',
    'script_container',
]


from typing import Any, List, Union, Sequence

from boa3.builtin.interop.contract.callflagstype import CallFlags
from boa3.builtin.interop.runtime.notification import Notification
from boa3.builtin.interop.runtime.triggertype import TriggerType
from boa3.builtin.type import ECPoint, UInt160


def check_witness(hash_or_pubkey: Union[UInt160, ECPoint]) -> bool:
    """
    Verifies that the transactions or block of the calling contract has validated the required script hash.

    >>> check_witness(calling_script_hash)
    True

    >>> check_witness(UInt160(bytes(20)))
    False

    :param hash_or_pubkey: script hash or public key to validate
    :type hash_or_pubkey: UInt160 or ECPoint
    :return: a boolean value that represents whether the script hash was verified
    :rtype: bool
    """
    pass


def notify(state: Any, notification_name: str = None):
    """
    Notifies the client from the executing smart contract.

    >>> var = 10
    ... notify(var)     # An event will be triggered
    None

    >>> var = 10
    ... notify(var, 'custom event name')     # An event will be triggered
    None

    :param state: the notification message
    :type state: Any
    :param notification_name: name that'll be linked to the notification
    :type notification_name: str
    """
    pass


def log(message: str):
    """
    Show log messages to the client from the executing smart contract.

    >>> log('log sent')     # An event that can be shown on the CLI
    log sent

    :param message: the log message
    :type message: str
    """
    pass


def get_trigger() -> TriggerType:
    """
    Return the smart contract trigger type.

    >>> get_trigger()
    TriggerType.APPLICATION

    :return: a value that represents the contract trigger type
    :rtype: TriggerType
    """
    pass


def get_notifications(script_hash: UInt160 = UInt160()) -> List[Notification]:
    """
    This method gets current invocation notifications from specific 'script_hash'.

    >>> notify(1); notify(2); notify(3)
    ... get_notifications(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))
    [
        [
            b'8\\xfe\\x11\\n\\xff7J\\xb8}\\xe9x6@\\xea\\x0b\\x00\\xf1|\\x82v',
            'notify',
            [1]
        ],
        [
            b'8\\xfe\\x11\\n\\xff7J\\xb8}\\xe9x6@\\xea\\x0b\\x00\\xf1|\\x82v',
            'notify',
            [2]
        ],
        [
            b'8\\xfe\\x11\\n\\xff7J\\xb8}\\xe9x6@\\xea\\x0b\\x00\\xf1|\\x82v',
            'notify',
            [3]
        ]
    ]

    :param script_hash: must have 20 bytes, but if it's all zero 0000...0000 it refers to all existing notifications
        (like a * wildcard)
    :type script_hash: UInt160
    :return: It will return an array of all matched notifications
    :rtype: List[Notification]
    """
    pass


def get_network() -> int:
    """
    Gets the magic number of the current network.

    >>> get_network()
    860243278

    :return: the magic number of the current network
    :rtype: int
    """
    pass


def burn_gas(gas: int):
    """
    Burns GAS to benefit the NEO ecosystem.

    >>> burn_gas(1000)
    None

    :param gas: the amount of GAS that will be burned
    :type gas: int

    :raise Exception: raised if gas value is negative.
    """
    pass


def get_random() -> int:
    """
    Gets the next random number.

    >>> get_random()
    191320526825634396960813166838892720709

    >>> get_random()
    99083669484001682562631729023191545809

    >>> get_random()
    328056213623902365838800581788496514419

    :return: the next random number
    :rtype: int
    """
    pass


def load_script(script: bytes, args: Sequence = (), flags: CallFlags = CallFlags.NONE) -> Any:
    """
    Loads a script at runtime.

    >>> from typing import cast
    ... from boa3.builtin.vm import Opcode
    ... cast(int, load_script(Opcode.ADD, [10, 2]))
    12

    """
    pass


address_version: int = 0
"""
Gets the address version of the current network.

>>> address_version
53

:meta hide-value:
"""


executing_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the current context.

>>> executing_script_hash
b'^b]\\x90#\\xbf\\xcc\\x1f\\xd8\\x9e\\xe3\\xa4zd\\x14\\xa4\\xf0\\x96\\x9f`'

:meta hide-value:
"""

calling_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the calling contract.

>>> calling_script_hash
b'\\x05\\x7f\\xc2\\x9d\\xba\\xb2\\xc1x\\xf5\\x81\\x83\\xbf\\xcb\\x87/\\xc3!\\xca\\xe1\\xd0'

:meta hide-value:
"""

time: int = 0
"""
Gets the timestamp of the current block.

>>> time
1685395697108

:meta hide-value:
"""

gas_left: int = 0
"""
Gets the remaining GAS that can be spent in order to complete the execution.

>>> gas_left
1999015490

:meta hide-value:
"""

platform: str = ''
"""
Gets the name of the current platform.

>>> platform
'NEO'

:meta hide-value:
"""

invocation_counter: int = 0
"""
Gets the number of times the current contract has been called during the execution.

>>> invocation_counter
1

:meta hide-value:
"""

entry_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the entry context.

>>> entry_script_hash
b'\\tK\\xb31\\xa8\\x13\\x80`\\xad\\xf6\\xda\\xdf\\xc6R\\x9b\\xfdB\\xbf\\x83\\x8f'

:meta hide-value:
"""

script_container: Any = None
"""
Gets the current script container.

>>> script_container
[
    b'\\xf1y\\xc2\\xd6\\x1c\\xb6\\x98\\xa4\\xdc\\xf3\\xd67s\\xd7E\\xf0<;\\x98+\\xa2T\\x03P,T\\xe8\\xc6{ \\x101', 
    0, 
    442907905, 
    '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00', 
    0, 
    0, 
    0, 
    ''
]

>>> script_container
[
    b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04",
    0,
    '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
    '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
    1468595301000,
    2083236893,
    0,
    0,
    b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
    0,
]

:meta hide-value:
"""



__all__ = [
    'CheckpointCreateCommand',
    'CheckpointRestoreCommand',
]

from boa3_test.test_drive.neoxp.command.neoexpresscommand.checkpoint.create import CheckpointCreateCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.checkpoint.restore import CheckpointRestoreCommand



from typing import Any, Dict, List, Optional, Sized, Tuple

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.annotation.metatype import MetaType, metaType
from boa3.internal.model.type.anytype import anyType
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class CastTypeMethod(IBuiltinMethod):

    def __init__(self, cast_to_type: IType = None, origin_type: IType = None):
        # if the type to be casted is not value, there's an error in the signature
        if cast_to_type is None:
            type_value = metaType
        elif isinstance(cast_to_type, MetaType):
            type_value = cast_to_type if cast_to_type.has_meta_type else metaType
        else:
            type_value = MetaType.build(cast_to_type)

        if origin_type is None:
            origin_type = anyType
        elif isinstance(origin_type, MetaType):
            origin_type = origin_type.meta_type if origin_type.has_meta_type else anyType

        identifier = 'cast'
        args: Dict[str, Variable] = {'typ': Variable(type_value),
                                     'val': Variable(anyType)}
        super().__init__(identifier, args,
                         return_type=cast_to_type.meta_type if isinstance(cast_to_type, MetaType) else anyType)
        self._origin_type: IType = origin_type

    @property
    def identifier(self) -> str:
        if self.is_supported:
            identifier = (self.typ_arg.type.meta_id
                          if hasattr(self.typ_arg.type, 'meta_id')
                          else self.typ_arg.type.identifier)
            return '-{0}_{1}'.format(self._identifier, identifier)
        else:
            return self._identifier

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def is_supported(self) -> bool:
        return (isinstance(self.typ_arg, Variable)
                and isinstance(self.typ_arg.type, MetaType)
                and self.typ_arg.type.has_meta_type)

    def not_supported_str(self, callable_id: str) -> str:
        return '{0}({1})'.format(callable_id,
                                 ','.join([arg.type.identifier if isinstance(arg, Variable) else 'unknown'
                                           for arg in self.args.values()]))

    @property
    def is_cast(self) -> bool:
        return True

    @property
    def cast_types(self) -> Optional[Tuple[IType, IType]]:
        origin_type = self._origin_type
        target_type = (self.typ_arg.type.meta_type
                       if hasattr(self.typ_arg.type, 'meta_type')
                       else self.typ_arg.type)
        return origin_type, target_type

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def _args_on_stack(self) -> int:
        return 1  # the implementation is the same as x = arg

    @property
    def args_on_stack(self) -> int:
        return 1  # the implementation is the same as x = arg

    @property
    def generation_order(self) -> List[int]:
        # type should not be converted
        indexes = super().generation_order
        typ_index = list(self.args).index('typ')

        if typ_index in indexes:
            indexes.remove(typ_index)

        return indexes

    @property
    def _body(self) -> Optional[str]:
        return None

    @property
    def typ_arg(self) -> Variable:
        return self.args['typ'] if isinstance(self.args['typ'], Variable) else Variable(anyType)

    @property
    def val_arg(self) -> Variable:
        return self.args['val']

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sized) and len(value) == len(self.args):
            cast_to = value[0]
            cast_from = value[1]

            if isinstance(cast_to, IType):
                if isinstance(cast_from, IType):
                    return CastTypeMethod(cast_to, cast_from)
                else:
                    return CastTypeMethod(cast_to)
        return super().build(value)



import asyncio
import unittest


class AsyncTestSuite(unittest.TestSuite):
    def run(self, result, debug=False):
        top_level = False
        if getattr(result, '_testRunEntered', False) is False:
            result._testRunEntered = top_level = True
        async_method = []
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        for index, test in enumerate(self):
            async_method.append(self.startRunCase(index, test, result))
        if async_method:
            loop.run_until_complete(asyncio.wait(async_method))
        loop.close()
        if top_level:
            self._tearDownPreviousClass(None, result)
            self._handleModuleTearDown(result)
            result._testRunEntered = False
        return result

    async def startRunCase(self, index, test, result):
        def _isnotsuite(test):
            try:
                iter(test)
            except TypeError:
                return True
            return False

        loop = asyncio.get_event_loop()
        if result.shouldStop:
            return False

        if _isnotsuite(test):
            self._tearDownPreviousClass(test, result)
            self._handleModuleFixture(test, result)
            self._handleClassSetUp(test, result)
            result._previousTestClass = test.__class__

            if (getattr(test.__class__, '_classSetupFailed', False) or
                    getattr(result, '_moduleSetUpFailed', False)):
                return True

        await loop.run_in_executor(None, test, result)

        if self._cleanup:
            self._removeTestAtIndex(index)


class CustomTestResult(unittest.TextTestResult):
    def startTest(self, test):
        super(unittest.TextTestResult, self).startTest(test)

    def addSuccess(self, test):
        self._write_description(test)
        super(CustomTestResult, self).addSuccess(test)

    def addError(self, test, err):
        self._write_description(test)
        super(CustomTestResult, self).addError(test, err)

    def addFailure(self, test, err):
        self._write_description(test)
        super(CustomTestResult, self).addFailure(test, err)

    def addSkip(self, test, reason):
        self._write_description(test)
        super(CustomTestResult, self).addSkip(test, reason)

    def addExpectedFailure(self, test, err):
        self._write_description(test)
        super(CustomTestResult, self).addExpectedFailure(test, err)

    def addUnexpectedSuccess(self, test):
        self._write_description(test)
        super(CustomTestResult, self).addUnexpectedSuccess(test)

    def _write_description(self, test):
        if self.showAll:
            self.stream.write(self.getDescription(test))
            self.stream.write(" ... ")
            self.stream.flush()


def list_of_tests_gen(test_suite):
    for test in test_suite:
        if hasattr(unittest.suite, '_isnotsuite') and unittest.suite._isnotsuite(test):
            yield test
        else:
            for t in list_of_tests_gen(test):
                yield t



import logging
import os
import sys
from argparse import _SubParsersAction
from typing import Optional

from boa3.boa3 import Boa3
from boa3.internal.cli_commands.icommand import ICommand
from boa3.internal.exception.NotLoadedException import NotLoadedException


class CompileCommand(ICommand):

    def __init__(self, main_parser: _SubParsersAction):
        super().__init__(main_parser, 'compile', 'Compiles your smart contract')

    def add_arguments_and_callback(self):
        self.parser.add_argument("input",
                                 type=str,
                                 help=".py smart contract to compile")
        self.parser.add_argument("-db", "--debug",
                                 action='store_true',
                                 help="generates a .nefdbgnfo file")
        self.parser.add_argument("--project-path",
                                 type=str,
                                 help="Project root path. Path of the contract by default.")
        self.parser.add_argument("-e", "--env",
                                 type=str,
                                 help="Set the contract environment for compiling.")
        self.parser.add_argument("-o", "--output-path",
                                 metavar='NEF_OUTPUT',
                                 type=str,
                                 default=None,
                                 help="Chooses the name and where the compiled files will be generated, "
                                      "if not specified it will be generated on the same directory with the same name "
                                      "as the python file.")
        self.parser.add_argument("--no-failfast",
                                 action='store_true',
                                 help="Do not stop on first compile error")
        self.parser.add_argument("--log-level",
                                 type=str,
                                 help="Log output level")

        self.parser.set_defaults(func=self.execute_command)

    @staticmethod
    def execute_command(args: dict):
        sc_path: str = args['input']
        project_path: str = args['project_path']
        debug: bool = args['debug']
        env: str = args['env']
        output_path: Optional[str] = args['output_path']
        fail_fast: bool = not args['no_failfast']
        log_level = args['log_level']

        if not sc_path.endswith(".py") or not os.path.isfile(sc_path):
            logging.error("Input file is not .py")
            sys.exit(1)

        fullpath = os.path.realpath(sc_path)
        path, filename = os.path.split(fullpath)

        if isinstance(output_path, str):
            if not output_path.endswith('.nef'):
                logging.error("Output path file extension is not .nef")
                sys.exit(1)

            path, filename = os.path.split(os.path.realpath(output_path))

        try:
            Boa3.compile_and_save(sc_path,
                                  output_path=output_path,
                                  debug=debug,
                                  root_folder=project_path,
                                  env=env,
                                  fail_fast=fail_fast,
                                  show_errors=True,
                                  log_level=log_level
                                  )
            logging.info(f"Wrote {filename.replace('.py', '.nef')} to {path}")
        except NotLoadedException as e:
            error_message = e.message
            log_error = 'Could not compile'
            if len(error_message) > 0:
                log_error += f': {error_message}'

            logging.error(log_error)
        except Exception as e:
            logging.exception(e)



from __future__ import annotations

import enum


class TriggerType(enum.IntEnum):
    # Indicate that the contract is triggered by the system to execute the OnPersist method of the native contracts.
    OnPersist = 0x01

    # Indicate that the contract is triggered by the system to execute the PostPersist method of the native contracts.
    PostPersist = 0x02

    # Indicates that the contract is triggered by the verification of a IVerifiable.
    Verification = 0x20

    # Indicates that the contract is triggered by the execution of transactions.
    Application = 0x40

    # The combination of all system triggers.
    System = OnPersist | PostPersist

    # The combination of all triggers.
    All = OnPersist | PostPersist | Verification | Application

    def neo_name(self) -> str:
        return self.name

    @staticmethod
    def get_from_neo_name(neo_name: str) -> TriggerType:
        return TriggerType[neo_name]



__all__ = [
    'get_exec_fee_factor',
    'get_fee_per_byte',
    'get_storage_price',
    'is_blocked',
]


from boa3.builtin.type import UInt160


def get_exec_fee_factor() -> int:
    """
    Gets the execution fee factor. This is a multiplier that can be adjusted by the committee to adjust the system fees
    for transactions.

    >>> get_exec_fee_factor()
    30

    :return: the execution fee factor
    :rtype: int
    """
    pass


def get_fee_per_byte() -> int:
    """
    Gets the network fee per transaction byte.

    >>> get_fee_per_byte()
    1000

    :return: the network fee per transaction byte
    :rtype: int
    """
    pass


def get_storage_price() -> int:
    """
    Gets the storage price.

    >>> get_storage_price()
    100000

    :return: the snapshot used to read data
    :rtype: int
    """
    pass


def is_blocked(account: UInt160) -> bool:
    """
    Determines whether the specified account is blocked.

    >>> is_blocked(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))
    False

    :param account: the account to be checked
    :type account: UInt160

    :return: whether the account is blocked or not
    :rtype: bool
    """
    pass



__all__ = ['OracleResponseCodeType',
           'OracleClass',
           ]

from boa3.internal.model.builtin.interop.oracle.oracleresponsecodetype import OracleResponseCodeType
from boa3.internal.model.builtin.native.oracleclass import OracleClass



from boa3_test.test_drive.neoxp.command.neoexpresscommand import (batch,
                                                                  contract,
                                                                  create,
                                                                  checkpoint,
                                                                  fastforward,
                                                                  reset,
                                                                  show,
                                                                  transfer,
                                                                  )
from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand



import ast
from typing import Dict, List, Tuple, Optional

from boa3.internal.compiler.compiledmetadata import CompiledMetadata
from boa3.internal.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode
from boa3.internal.neo.vm.opcode.OpcodeInfo import OpcodeInfo


class NativeContractMethod(InteropMethod):

    def __init__(self, native_contract_script_hash_method: ContractGetHashMethod, identifier: str, syscall: str,
                 args: Dict[str, Variable] = None, defaults: List[ast.AST] = None, return_type: IType = None,
                 internal_call_args: int = None):
        super().__init__(identifier, syscall, args, defaults, return_type)
        self.script_hash_method = native_contract_script_hash_method
        min_no_args = len(self.args_without_default)
        if not isinstance(internal_call_args, int):
            internal_call_args = len(self.args)
        elif len(self.args_without_default) > internal_call_args:
            internal_call_args = min_no_args

        self.internal_call_args = internal_call_args

        from boa3.internal.neo3.contracts.contracttypes import CallFlags
        self._call_flags_default = CallFlags.ALL
        self._pack_arguments = None   # defined during compilation
        self._method_token_id = None  # defined during compilation
        self.external_name = self._sys_call

    @property
    def contract_script_hash(self) -> bytes:
        return self.script_hash_method.script_hash

    def reset(self):
        # reset the object state to ensure the correct output when calling consecutive compilations
        super().reset()
        self._pack_arguments = None
        self._method_token_id = None

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.Integer import Integer
        from boa3.internal.neo.vm.type.String import String

        call_flag = self._call_flags_default
        if self._pack_arguments is None:
            self._pack_arguments = False

        CompiledMetadata.instance().add_contract_permission(self.contract_script_hash, self._sys_call)
        self_method_token_id = self._get_method_token_id(call_flag)

        if isinstance(self_method_token_id, int) and self_method_token_id >= 0:
            call_opcode = Opcode.CALLT
            opcode_info = OpcodeInfo.get_info(call_opcode)
            arg_size = opcode_info.data_len

            call_opcodes = [
                (call_opcode, Integer(self_method_token_id).to_byte_array(min_length=arg_size))
            ]
        else:
            if len(self.args) == 0:
                pack_args = [(Opcode.NEWARRAY0, b'')]
            else:
                from boa3.internal.neo.vm.opcode import OpcodeHelper
                pack_args = [
                    OpcodeHelper.get_push_and_data(len(self.args)),
                    (Opcode.PACK, b'')
                ]

            call_flags = Integer(call_flag).to_byte_array(signed=True, min_length=1)
            flags_opcode = [
                (Opcode.PUSHDATA1, Integer(len(call_flags)).to_byte_array() + call_flags)
            ]

            method = String(self._sys_call).to_bytes()
            method_opcode = [
                (Opcode.PUSHDATA1, Integer(len(method)).to_byte_array(min_length=1) + method)
            ]

            drop_if_void_opcode = [
                (Opcode.DROP, b'')
            ] if self.return_type is Type.none else []

            call_opcodes = (pack_args
                            + flags_opcode
                            + method_opcode
                            + self.script_hash_method.opcode
                            + Interop.CallContract.opcode
                            + drop_if_void_opcode
                            )

        return call_opcodes

    def _get_method_token_id(self, call_flag=None) -> Optional[int]:
        if self._method_token_id is None:
            from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping

            if call_flag is None:
                call_flag = self._call_flags_default
            self._method_token_id = VMCodeMapping.instance().add_method_token(self, call_flag)
        return self._method_token_id

    @property
    def pack_arguments(self) -> bool:
        if self._pack_arguments is None:
            from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
            self_method_token_id = self._get_method_token_id(self._call_flags_default)

            self._pack_arguments = not isinstance(self_method_token_id, int) or self_method_token_id < 0
        return self._pack_arguments

    @property
    def method_name(self) -> str:
        return self._sys_call



from typing import Dict, List, Optional, Tuple

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class SqrtMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'sqrt'
        args: Dict[str, Variable] = {'val': Variable(Type.int)}
        super().__init__(identifier, args, return_type=Type.int)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.SQRT, b'')]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



from typing import Any, Dict, List, Optional, Tuple

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.collection.sequence.tupletype import TupleType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode import OpcodeHelper
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class MinMethod(IBuiltinMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.internal.model.type.type import Type
        identifier = 'min'

        self._allowed_types = [Type.int, Type.str, Type.bytes]
        default_type = Type.int
        if not self._is_valid_type(arg_value):
            arg_value = default_type

        args: Dict[str, Variable] = {
            'args1': Variable(arg_value),
            'args2': Variable(arg_value)
        }
        vararg = ('values', Variable(arg_value))
        super().__init__(identifier, args, return_type=arg_value, vararg=vararg)

    def _is_valid_type(self, arg_type: Optional[IType]) -> bool:
        return (isinstance(arg_type, IType) and
                any(allowed_type.is_type_of(arg_type) for allowed_type in self._allowed_types))

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_values.type is Type.int:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_values.type._identifier)

    @property
    def _arg_values(self) -> Variable:
        return self._vararg[1]

    @property
    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.compiler.codegenerator import get_bytes_count
        from boa3.internal.neo.vm.type.Integer import Integer

        jmp_place_holder = (Opcode.JMP, b'\x01')

        verify_number_of_parameters = [  # verifies if the stack has 2 or 3 items
            (Opcode.DEPTH, b''),
            (Opcode.PUSH2, b''),
            (Opcode.JMPEQ, b''),
            jmp_place_holder
        ]

        if_n_parameters_gt_2 = [  # if number of items in stack is 3
            (Opcode.REVERSE3, b''),
            (Opcode.UNPACK, b''),
            (Opcode.INC, b''),
            (Opcode.INC, b''),
            jmp_place_holder  # skips the next block of instructions
        ]

        if_n_parameters_eq_2 = [  # if number of items in stack is 2
            (Opcode.PUSH2, b'')
        ]

        jmp_n_parameters_eq_2 = OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(if_n_parameters_eq_2), True)
        if_n_parameters_gt_2[-1] = jmp_n_parameters_eq_2

        jmp_n_parameters_gt_2 = OpcodeHelper.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(if_n_parameters_gt_2))
        verify_number_of_parameters[-1] = jmp_n_parameters_gt_2

        repack_array = [  # pack all the arguments in the array
            (Opcode.PACK, b''),
        ]

        is_int_initialize = [  # puts the last array element as the max value
            (Opcode.DUP, b''),  # index = len(array) - 1
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),     # min = array[index]
        ]

        is_int_while = [    # this will get the next number in the array and compare it with the current min
            (Opcode.SWAP, b''),         # index--
            (Opcode.DEC, b''),
            (Opcode.SWAP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),         # min = min if min < array[index] else array[index]
            (Opcode.PICKITEM, b''),
        ]
        is_int_while.extend(self._compare_values())
        is_int_while.extend([
            (Opcode.OVER, b''),
            (Opcode.SIGN, b'')
            # if index != 0: go back to index--
            # else go to the end
        ])

        jmp_back_to_while_statement = (Opcode.JMPIF, Integer(-get_bytes_count(is_int_while)).to_byte_array(signed=True))
        is_int_while.append(jmp_back_to_while_statement)

        clean_stack = [    # removes everything but min
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        return (
            verify_number_of_parameters +
            if_n_parameters_gt_2 +
            if_n_parameters_eq_2 +
            repack_array +
            is_int_initialize +
            is_int_while +
            clean_stack
        )

    def _compare_values(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.MIN, b'')]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) > 0:
            value = value[0]
        if isinstance(value, TupleType):
            value = value.value_type
        if type(value) == type(self._arg_values.type):
            return self

        from boa3.internal.model.builtin.method.minbytestringmethod import MinByteStringMethod
        from boa3.internal.model.builtin.method.minintmethod import MinIntMethod
        from boa3.internal.model.type.type import Type

        if Type.str.is_type_of(value) or Type.bytes.is_type_of(value):
            return MinByteStringMethod(value)

        return MinIntMethod(value)



__all__ = ['ContractManagementClass',
           'CryptoLibClass',
           'GasClass',
           'LedgerClass',
           'NeoClass',
           'PolicyClass',
           'RoleManagementClass',
           'StdLibClass',
           ]

from boa3.internal.model.builtin.native.contractmanagementclass import ContractManagementClass
from boa3.internal.model.builtin.native.cryptolibclass import CryptoLibClass
from boa3.internal.model.builtin.native.gasclass import GasClass
from boa3.internal.model.builtin.native.ledgerclass import LedgerClass
from boa3.internal.model.builtin.native.neoclass import NeoClass
from boa3.internal.model.builtin.native.policyclass import PolicyClass
from boa3.internal.model.builtin.native.rolemanagementclass import RoleManagementClass
from boa3.internal.model.builtin.native.stdlibclass import StdLibClass



import ast
from typing import Dict, List

from boa3.internal.model.builtin.interop.nativecontract.Neo.getneoscripthashmethod import NeoContract
from boa3.internal.model.builtin.interop.nativecontract.nativecontractmethod import NativeContractMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class NeoContractMethod(NativeContractMethod):

    def __init__(self, identifier: str, native_identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None,
                 internal_call_args: int = None):
        super().__init__(NeoContract.getter, identifier, native_identifier,
                         args, defaults, return_type, internal_call_args)



from typing import Any, Dict, List, Optional, Tuple

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.collection.sequence.tupletype import TupleType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintMethod(IBuiltinMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.internal.model.type.type import Type
        identifier = 'print'
        if not isinstance(arg_value, IType):
            arg_value = Type.str

        args: Dict[str, Variable] = {}
        vararg = ('values', Variable(arg_value))
        super().__init__(identifier, args, return_type=Type.none, vararg=vararg)

        self._print_value_opcodes = None

    @property
    def _arg_values(self) -> Variable:
        return self._vararg[1]

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_values.type is Type.str:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_values.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.compiler.codegenerator import get_bytes_count
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.neo.vm.type.Integer import Integer

        check_if_arg_is_empty = [
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSH0, b''),
        ]
        copy_list_and_reverse = [
            (Opcode.UNPACK, b''),  # copy - it must not change the original
            (Opcode.PACK, b''),
            (Opcode.DUP, b''),
            (Opcode.REVERSEITEMS, b'')
        ]

        complete_loop = (
            [
                (Opcode.DUP, b''),
                (Opcode.POPITEM, b''),
            ] + self.print_value_opcodes
            + Interop.Log.opcode
            + check_if_arg_is_empty
        )
        complete_loop.append((Opcode.JMPNE, Integer(-get_bytes_count(complete_loop)).to_byte_array(signed=True)))

        return (check_if_arg_is_empty +
                [
                    (Opcode.JMPEQ, Integer(get_bytes_count(copy_list_and_reverse)
                                           + 1  # the size of the JMPEQ arg
                                           + get_bytes_count(complete_loop)).to_byte_array(signed=True)),
                ] +
                copy_list_and_reverse +
                complete_loop +
                [
                    (Opcode.DROP, b'')
                ])

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:
            self._print_value_opcodes = []

        return self._print_value_opcodes

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        if isinstance(value, TupleType):
            value = value.value_type
        if type(value) == type(self._arg_values.type):
            return self

        from boa3.internal.model.builtin.method.printbytestringmethod import PrintByteStringMethod
        from boa3.internal.model.type.classes.userclass import UserClass
        from boa3.internal.model.type.type import Type

        if Type.bool.is_type_of(value):
            from boa3.internal.model.builtin.method.printboolmethod import PrintBoolMethod
            return PrintBoolMethod()

        elif Type.int.is_type_of(value):
            from boa3.internal.model.builtin.method.printintmethod import PrintIntMethod
            return PrintIntMethod()

        elif Type.str.is_type_of(value) or Type.bytes.is_type_of(value):
            return PrintByteStringMethod(value)

        elif isinstance(value, UserClass):
            from boa3.internal.model.builtin.method.printclassmethod import PrintClassMethod
            return PrintClassMethod(value)

        elif Type.sequence.is_type_of(value):
            from boa3.internal.model.builtin.method.printsequencemethod import PrintSequenceMethod
            return PrintSequenceMethod(value)

        return PrintByteStringMethod(value)



import ast
from typing import Dict, List

from boa3.internal.model.builtin.interop.nativecontract.ContractManagement.getcontractmanagementscripthashmethod import \
    ContractManagement
from boa3.internal.model.builtin.interop.nativecontract.nativecontractmethod import NativeContractMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class ContractManagementMethod(NativeContractMethod):

    def __init__(self, identifier: str, syscall: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None,
                 internal_call_args: int = None):
        super().__init__(ContractManagement.getter, identifier, syscall,
                         args, defaults, return_type, internal_call_args)



from typing import List, Tuple

from boa3.internal.model.builtin.method.printmethod import PrintMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintSequenceMethod(PrintMethod):

    def __init__(self, arg_value: IType):
        from boa3.internal.model.type.type import Type

        if not isinstance(arg_value, IType) or not Type.sequence.is_type_of(arg_value):
            arg_value = Type.list

        super().__init__(arg_value)

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:
            from boa3.internal.model.builtin.interop.interop import Interop
            self._print_value_opcodes = Interop.JsonSerialize.opcode.copy()

        return super().print_value_opcodes



import abc

from boa3.internal.neo3.core.types import UInt256


class ITransactionObject(abc.ABC):
    tx_hash_group = 'txhash'

    def __init__(self):
        self._log: str = ''
        self._tx_id: UInt256 = None

    def set_log(self, log: str):
        if not self._log:
            self._log = log

    @property
    def tx_id(self) -> UInt256:
        if not isinstance(self._tx_id, UInt256) and not self._log.isspace():
            self._handle_log()

        return self._tx_id

    def _log_pattern(self) -> str:
        return rf'(?P<{self.tx_hash_group}>0x\w+) submitted$'

    def _match_log(self) -> dict:
        import re
        pattern = self._log_pattern()
        if not pattern.startswith('^'):
            pattern = r'.*?' + pattern
        return re.match(pattern, self._log).groupdict()

    def _set_data_from_match_result(self, match_groups: dict):
        tx_hash = match_groups[self.tx_hash_group]
        self._tx_id = UInt256.from_string(tx_hash)

    def _handle_log(self):
        try:
            match_groups = self._match_log()
            self._set_data_from_match_result(match_groups)
        except:
            pass



from typing import List, Optional, Tuple

from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.collection.icollection import ICollectionType
from boa3.internal.model.type.collection.mapping.mappingtype import MappingType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class CollectionMembership(BinaryOperation):
    """
    A class used to represent a collection membership operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.collection]

    def __init__(self, left: IType = Type.any, right: IType = None):
        self.operator: Operator = Operator.In

        if not isinstance(right, ICollectionType):
            right = Type.collection

        reference_type = right.key_type if isinstance(right, MappingType) else right.item_type
        if not reference_type.is_type_of(left) or (reference_type is Type.any and left is not Type.any):
            if Type.sequence.is_type_of(left):
                right = Type.collection.build(left)
            else:
                right = Type.collection.build(Type.list.build([left]))

        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        if not isinstance(right, ICollectionType):
            return False

        if isinstance(right, PrimitiveType) and (right.is_type_of(left) or left.is_type_of(right)):
            return True

        if isinstance(right, MappingType):
            reference_type = right.key_type
        else:
            reference_type = right.item_type

        return reference_type.is_type_of(left)

    def _get_result(self, left: IType, right: IType) -> IType:
        return Type.bool

    def get_valid_operand_for_validation(self, left_operand: IType,
                                         right_operand: IType = None) -> Tuple[Optional[IType], Optional[IType]]:
        if isinstance(right_operand, ICollectionType):
            left = right_operand.item_type if not isinstance(right_operand, MappingType) else right_operand.key_type
            return left, right_operand

        return left_operand, self.right_type

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.neo.vm.type.Integer import Integer
        return [
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, Type.mapping.stack_item),   # if isinstance(arg1, dict)
            (Opcode.JMPIFNOT, Integer(6).to_byte_array(signed=True, min_length=1)),
            (Opcode.SWAP, b''),
            (Opcode.HASKEY, b''),                           # return value.has_key(arg0)
            (Opcode.JMP, Integer(54).to_byte_array(signed=True, min_length=1)),
            (Opcode.DUP, b''),
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, Type.str.stack_item),       # is_bytestr = isinstance(arg1, (str, bytes))
            (Opcode.SWAP, b''),
            (Opcode.SIZE, b''),                         # limit = len(arg1)
            (Opcode.OVER, b''),                         # if is_bytestr:
            (Opcode.JMPIFNOT, Integer(7).to_byte_array(signed=True, min_length=1)),
            (Opcode.PUSH3, b''),                            # limit = limit - len(arg0) + 1
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.SUB, b''),
            (Opcode.INC, b''),
            (Opcode.PUSH0, b''),                        # index = 0
            (Opcode.DUP, b''),                          # while index < limit:
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.LT, b''),
            (Opcode.JMPIFNOT, Integer(28).to_byte_array(signed=True, min_length=1)),
            (Opcode.PUSH4, b''),                            # aux = arg0
            (Opcode.PICK, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH5, b''),
            (Opcode.PICK, b''),                             # if is_bytestr:
            (Opcode.JMPIFNOT, Integer(10).to_byte_array(signed=True, min_length=1)),
            (Opcode.PUSH2, b''),                                # if arg1[index:len(arg0)] == arg0:
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, Type.str.stack_item),
            (Opcode.NUMEQUAL, b''),                                 # break
            (Opcode.JMP, Integer(4).to_byte_array(signed=True, min_length=1)),
            (Opcode.PICKITEM, b''),                         # elif arg1[index] == arg0:
            (Opcode.EQUAL, b''),                                    # break
            (Opcode.JMPIF, Integer(5).to_byte_array(signed=True, min_length=1)),
            (Opcode.INC, b''),                              # index += 1
            (Opcode.JMP, Integer(-30).to_byte_array(signed=True, min_length=1)),
            (Opcode.GT, b''),                       # return index < limit
            (Opcode.REVERSE4, b''),                     # if the value is found, index won't reach the limit
            (Opcode.DROP, b''),                     # remove auxiliar values from stack
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]



from __future__ import annotations

from typing import List, Dict, Any, Union

from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.model.network.payloads.signer import Signer
from boa3_test.test_drive.model.network.payloads.testtransaction import TestTransaction
from boa3_test.test_drive.model.network.payloads.witness import Witness
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp import utils


class TestRunnerTransaction(TestTransaction):
    def __init__(self, tx_hash: Union[UInt256, bytes], script: bytes, sender: Account):
        super().__init__(tx_hash, script)
        self._size = 0
        self._version = 0
        self._nonce = 0
        self._sender = sender
        self._system_fee = 0
        self._network_fee = 0
        self._valid_until_block = 0

    @property
    def script(self) -> bytes:
        return self._script

    @property
    def size(self) -> int:
        return self._size

    @property
    def version(self) -> int:
        return self._version

    @property
    def nonce(self) -> int:
        return self._nonce

    @property
    def sender(self) -> Account:
        return self._sender

    @property
    def system_fee(self) -> int:
        return self._system_fee

    @property
    def network_fee(self) -> int:
        return self._network_fee

    @property
    def valid_until_block(self) -> int:
        return self._valid_until_block

    @property
    def signers(self) -> List[Signer]:
        return self._signers.copy()

    @property
    def attributes(self) -> list:
        return self._attributes.copy()

    @property
    def witnesses(self) -> List[Witness]:
        return self._witnesses.copy()

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> TestRunnerTransaction:
        tx: TestRunnerTransaction = super().from_json(json)

        tx._size = json['size']
        tx._version = json['version']
        tx._nonce = json['nonce']
        sender_account = json['sender']
        tx._sender = utils.get_account_from_script_hash_or_id(sender_account)
        tx._system_fee = int(json['sysfee'])
        tx._network_fee = int(json['netfee'])
        tx._valid_until_block = json['validuntilblock']

        return tx



import threading
from typing import List, Union, Any, Type

from boa3.internal.neo.vm.type.String import String
from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.interface.itransactionobject import ITransactionObject
from boa3_test.test_drive.model.invoker.neobatchinvoke import NeoBatchInvoke
from boa3_test.test_drive.model.invoker.neoinvoke import NeoInvoke
from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope
from boa3_test.test_drive.model.smart_contract.testcontract import TestContract
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp import utils
from boa3_test.test_drive.neoxp.command import neoxp
from boa3_test.test_drive.neoxp.command.neoexpresscommand import NeoExpressCommand

_NEOXP_BATCH_LOCK = threading.Lock()


class NeoExpressBatch:
    def __init__(self):
        self._instructions: List[NeoExpressCommand] = []
        self._transaction_submissions: List[int] = []
        self._transaction_logs: List[str] = []

        self._tx_invokes: List[ITransactionObject] = []
        self._tx_invokes_pos: List[int] = []

    def is_empty(self) -> bool:
        return len(self._instructions) == 0

    def cur_size(self) -> int:
        return len(self._instructions)

    def create_neo_express_checkpoint(self, checkpoint_path: str, overwrite: bool = False):
        self._instructions.append(neoxp.create_checkpoint(checkpoint_path, force=overwrite))

    def deploy_contract(self, nef_path: str, deployer: Account = None) -> TestContract:
        if hasattr(deployer, 'name') and isinstance(deployer.name, str):
            deployer_id = deployer
        else:
            deployer_id = utils.get_default_account()  # neo express default account

        tx_pos = len(self._instructions)
        deployed_contract = TestContract(nef_path, nef_path.replace('.nef', '.manifest.json'))

        self._instructions.append(neoxp.contract.deploy(nef_path, deployer_id))
        self._transaction_submissions.append(tx_pos)

        self._tx_invokes.append(deployed_contract)
        self._tx_invokes_pos.append(tx_pos)

        return deployed_contract

    def run_contract(self, invoke: NeoInvoke, witness_scope: WitnessScope = None,
                     expected_result_type: Type = None) -> NeoBatchInvoke:
        if hasattr(invoke.invoker, 'name') and isinstance(invoke.invoker.name, str):
            invoker_id = invoke.invoker
        else:
            invoker_id = utils.get_default_account()  # neo express default account

        tx_pos = len(self._instructions)
        batch_invoke = NeoBatchInvoke(invoke, tx_pos, expected_result_type=expected_result_type)

        self._instructions.append(neoxp.contract.run(invoke.contract_id, invoke.operation, *invoke.cli_args,
                                                     account=invoker_id, witness_scope=witness_scope))
        self._transaction_submissions.append(tx_pos)

        self._tx_invokes.append(batch_invoke)
        self._tx_invokes_pos.append(tx_pos)

        return batch_invoke

    def invoke_file(self, invoke_file: str, account: Account = None, witness_scope: WitnessScope = None):
        if hasattr(account, 'name') and isinstance(account.name, str):
            invoker_id = account
        else:
            invoker_id = utils.get_default_account()  # neo express default account

        if witness_scope != WitnessScope.CalledByEntry:
            # only CalledByEntry and Global are accepted as WitnessScopes in neo express
            witness_scope = WitnessScope.Global

        self._instructions.append(neoxp.contract.invoke(invoke_file, invoker_id, witness_scope))

    def transfer_assets(self, sender: Account, receiver: Account, asset: Union[str, UInt160],
                        quantity: Union[int, float], decimals: int = 0, data: Any = None):
        if isinstance(asset, UInt160):
            asset = asset.__str__()

        self._instructions.append(neoxp.transfer(sender, receiver, asset,
                                                 quantity, decimals=decimals,
                                                 data=data)
                                  )

    def mint_block(self, block_count: int, time_interval_in_secs: int = 0):
        self._instructions.append(neoxp.fastfwd(block_count, time_interval_in_secs))

    def reset_blockchain(self):
        self._instructions.append(neoxp.reset())

    def clear(self):
        self._instructions.clear()
        self._transaction_submissions.clear()
        self._transaction_logs.clear()
        self._tx_invokes.clear()
        self._tx_invokes_pos.clear()

    def pop(self):
        result = self._instructions.pop()

        pop_index = len(self._instructions)
        if pop_index in self._transaction_submissions:
            index = self._transaction_submissions.index(len(self._instructions))
            self._transaction_submissions.pop(index)
            if len(self._transaction_logs) > index:
                self._transaction_logs.pop(index)
        if pop_index in self._tx_invokes_pos:
            index = self._tx_invokes_pos.index(len(self._instructions))
            self._tx_invokes_pos.pop(index)
            self._tx_invokes.pop(index)

        return result

    def write(self, path: str) -> bool:
        try:
            with open(path, 'wb+') as json_file:
                for instruction in self._instructions:
                    json_file.write(String(f'{instruction.cli_command()}\n').to_bytes())

            return True
        except BaseException:
            return False

    def execute(self, neoxp_path: str, batch_file_path: str, reset: bool = False, check_point_file: str = None) -> str:
        with _NEOXP_BATCH_LOCK:
            self.write(batch_file_path)
            log = utils.run_batch(neoxp_path, batch_file_path, reset=reset, check_point_file=check_point_file)

        self._transaction_logs.clear()
        self._update_logs(log)
        self._remove_commands_until_checkpoint(check_point_file)
        return log

    def _update_logs(self, log: str):
        import re
        cli_color_change = r'\xb1?(?:\[\d+m)?'
        log_regex = rf'{cli_color_change}(?P<contract>.+?){cli_color_change}\sLog:\s{cli_color_change}'
        tx_logs = []

        # filter to remove logs
        logs = log.splitlines()
        max_log_count = len(self._instructions)

        index = len(logs) - 1
        while index >= 0 and 0 < len(logs) - index <= max_log_count:
            line = logs[index]
            if re.match(log_regex, line):
                logs.pop(index)
            index -= 1

        for lineno, line in enumerate(logs[-max_log_count:]):
            if lineno in self._transaction_submissions:
                tx_logs.append(line)
            if lineno in self._tx_invokes_pos:
                index = self._tx_invokes_pos.index(lineno)
                self._tx_invokes[index].set_log(line)

        self._transaction_logs = tx_logs

    def _remove_commands_until_checkpoint(self, check_point_file: str):
        if isinstance(check_point_file, str):
            condition = (index + 1 for index, command in enumerate(self._instructions)
                         if (isinstance(command, neoxp.neoxp.checkpoint.CheckpointCreateCommand)
                             and command.file_name == check_point_file
                             )
                         )
        else:
            condition = (index + 1 for index, command in enumerate(self._instructions)
                         if isinstance(command, neoxp.neoxp.checkpoint.CheckpointCreateCommand)
                         )

        instr_to_remove = next(condition, None)
        if isinstance(instr_to_remove, int):
            self._instructions[:instr_to_remove] = []

            last_tx_submitted = next((index for index, tx_pos in enumerate(self._transaction_submissions)
                                      if tx_pos > instr_to_remove
                                      ), len(self._transaction_submissions))
            last_sc_submitted = next((index for index, tx_pos in enumerate(self._tx_invokes_pos)
                                      if tx_pos > instr_to_remove
                                      ), len(self._tx_invokes))

            self._transaction_logs[:last_tx_submitted] = []
            self._transaction_submissions[:last_tx_submitted] = []
            self._tx_invokes[:last_sc_submitted] = []
            self._tx_invokes_pos[:last_sc_submitted] = []

    def has_new_deploys(self):
        return next((True for command in self._instructions
                     if isinstance(command, neoxp.neoxp.contract.deploy.ContractDeployCommand)),
                    False)



from typing import Optional

from boa3.internal import constants
from boa3.internal.neo import to_hex_str
from boa3_test.test_drive.model.interface.itransactionobject import ITransactionObject


class TestContract(ITransactionObject):
    contract_hash_group = 'scripthash'

    def __init__(self, file_path: str, manifest_path: str):
        super().__init__()
        self._nef_path: str = file_path
        self._manifest_path: str = manifest_path
        self._manifest: dict = self._read_manifest()
        self._script_hash: Optional[bytes] = None

    @property
    def path(self) -> str:
        return self._nef_path

    @property
    def script_hash(self) -> Optional[bytes]:
        if not isinstance(self._script_hash, bytes) and len(self._log) > 0 and not self._log.isspace():
            self._handle_log()
        return self._script_hash

    @script_hash.setter
    def script_hash(self, new_value: Optional[bytes]):
        if self._script_hash is None and isinstance(new_value, bytes) and len(new_value) == constants.SIZE_OF_INT160:
            self._script_hash = new_value

    @property
    def name(self) -> str:
        return self._manifest['name'] if 'name' in self._manifest else None

    def _log_pattern(self) -> str:
        return rf'{self.name} \((?P<{self.contract_hash_group}>0x\w+)\).*?' + super()._log_pattern()

    def _set_data_from_match_result(self, match_groups: dict):
        from boa3.internal.neo3.core.types import UInt160
        super()._set_data_from_match_result(match_groups)
        script_hash = match_groups[self.contract_hash_group]
        self._script_hash = UInt160.from_string(script_hash).to_array()

    def _read_manifest(self) -> dict:
        try:
            with open(self._manifest_path) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())
        except BaseException:
            manifest = {}

        return manifest

    def is_valid_identifier(self, item: str) -> bool:
        # valid identifiers are nef path or a representation of the script hash
        return (item == self._nef_path
                or item == self._script_hash
                or (self._script_hash is not None and item == to_hex_str(self._script_hash)))

    def __str__(self) -> str:
        return '[{0}] {1}'.format(to_hex_str(self._script_hash) if isinstance(self._script_hash, bytes) else ' ',
                                  self.name)

    def __repr__(self) -> str:
        return str(self)

    def get_identifier(self) -> str:
        if self.name.isspace() and self._script_hash is not None:
            return to_hex_str(self._script_hash)
        return self.name if ' ' not in self.name else f'"{self.name}"'



from typing import Dict, List

from boa3.internal.model.callable import Callable
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.annotation.metatype import metaType
from boa3.internal.model.type.type import Type
from boa3.internal.model.type.typingmethod.casttypemethod import CastTypeMethod


class TypeUtils:
    @classmethod
    def all_functions(cls) -> Dict[str, Callable]:
        from boa3.internal.model.builtin.builtincallable import IBuiltinCallable
        return {tpe._identifier: tpe for tpe in vars(cls).values() if isinstance(tpe, IBuiltinCallable)}

    @classmethod
    def get_types_from_typing_lib(cls) -> Dict[str, ISymbol]:
        import typing
        from types import FunctionType

        type_symbols: Dict[str, ISymbol] = {}
        all_types: List[str] = typing.__all__

        for t_id in all_types:
            attr = getattr(typing, t_id)
            if not isinstance(attr, FunctionType):
                type_id: str = t_id if t_id in Type.all_types() else t_id.lower()
                if type_id in Type.all_types():
                    type_symbols[t_id] = Type.all_types()[type_id]
            else:
                function_id: str = t_id if t_id in cls.all_functions() else t_id.lower()
                if function_id in cls.all_functions():
                    type_symbols[t_id] = cls.all_functions()[function_id]

        return type_symbols

    # type for internal validation
    type = metaType

    # Annotation function utils
    cast = CastTypeMethod()

    _internal_validation_symbols: List[IdentifiedSymbol] = [type
                                                            ]

    symbols_for_internal_validation = {symbol.identifier: symbol
                                       for symbol in _internal_validation_symbols}



import logging
import os.path
import re
import subprocess
import threading
from typing import Tuple, List, Union

from filelock import FileLock

from boa3.internal import env
from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.model.wallet import utils as wallet_utils
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import neoexpresscommand as neoxp
from boa3_test.test_drive.neoxp.model.neoxpconfig import NeoExpressConfig
from boa3_test.test_drive.testrunner.blockchain.block import TestRunnerBlock as Block
from boa3_test.test_drive.testrunner.blockchain.contract import TestRunnerContract as Contract
from boa3_test.test_drive.testrunner.blockchain.transaction import TestRunnerTransaction as Transaction
from boa3_test.test_drive.testrunner.blockchain.transactionlog import TestRunnerTransactionLog as TransactionLog

_NEOXP_CONFIG = NeoExpressConfig(f'{env.NEO_EXPRESS_INSTANCE_DIRECTORY}/default.neo-express')
_NEOXP_CONFIG_LOCK = threading.Lock()
_NEOXP_FILE_LOCK = FileLock(f'{env.TEST_RUNNER_DIRECTORY}/test-runner.lock')
logging.getLogger("filelock").setLevel(logging.INFO)


def get_account_by_name(account_name) -> Account:
    return next((account for account in _NEOXP_CONFIG.accounts if account.name == account_name),
                None)


def get_account_by_address(account_address: str) -> Account:
    return next((account for account in _NEOXP_CONFIG.accounts if account.address == account_address),
                None)


def get_account_by_identifier(account_identifier: str) -> Account:
    return next((account for account in _NEOXP_CONFIG.accounts if (account.address == account_identifier
                                                                   or account.name == account_identifier)),
                None)


def get_default_account() -> Account:
    return _NEOXP_CONFIG.default_account


def get_address_version() -> int:
    return _NEOXP_CONFIG.version


def get_magic() -> int:
    return _NEOXP_CONFIG.magic


def get_genesis_block() -> Block:
    return _NEOXP_CONFIG.genesis_block


def _set_genesis_block(found_block: Block):
    if _NEOXP_CONFIG.genesis_block is None:  # avoid overwrite
        _NEOXP_CONFIG._genesis_block = found_block


def get_account_from_script_hash_or_id(script_hash_or_address: Union[bytes, str]) -> Account:
    if isinstance(script_hash_or_address, bytes):
        script_hash = script_hash_or_address
        address = wallet_utils.address_from_script_hash(script_hash, get_address_version())
        account = get_account_by_address(address)
    elif isinstance(script_hash_or_address, str):
        account = get_account_by_identifier(script_hash_or_address)
        script_hash = wallet_utils.address_to_script_hash(script_hash_or_address, get_address_version())
    else:
        raise TypeError(f"Invalid data type {type(script_hash_or_address)}. Expecting str or bytes")

    if not isinstance(account, Account):
        from boa3.internal.neo3.core.types import UInt160
        from boa3_test.test_drive.neoxp.model.neoxpaccount import NeoExpressAccount
        account = NeoExpressAccount(UInt160(script_hash), get_address_version())

    return account


def create_neo_express_instance(neoxp_path: str) -> str:
    command = neoxp.create.CreateCommand(config_output=neoxp_path,
                                         node_count=1,
                                         force=True)
    with _NEOXP_CONFIG_LOCK:
        stdout, stderr = run_neo_express_cli(command)

    return stdout


def reset_neo_express_instance(neoxp_path: str, check_point_file: str = None) -> str:
    if isinstance(check_point_file, str) and os.path.isfile(check_point_file):
        command = neoxp.checkpoint.CheckpointRestoreCommand(neo_express_data_file=neoxp_path,
                                                            file_name=check_point_file,
                                                            force=True)
    else:
        command = neoxp.reset.ResetCommand(neo_express_data_file=neoxp_path,
                                           force=True)
    stdout, stderr = run_neo_express_cli(command)
    return stdout


def get_deployed_contracts(neoxp_path: str, check_point_file: str = None) -> List[Contract]:
    command = neoxp.contract.ContractListCommand(neo_express_data_file=neoxp_path)
    with _NEOXP_CONFIG_LOCK:
        reset_neo_express_instance(neoxp_path, check_point_file)
        stdout, stderr = run_neo_express_cli(command)

    contracts = []
    for line in stdout.splitlines():
        try:
            groups = re.match(r'(?P<name>.*?) \((?P<scripthash>0x\w+)\)', line).groupdict()
            contract_name = groups['name']
            contract_hash = groups['scripthash']
            found_contract = Contract(contract_name, contract_hash)
            contracts.append(found_contract)
        except:
            # don't break if some line doesn't match the regex
            continue
    return contracts


def _get_transaction_raw(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None) -> str:
    command = neoxp.show.ShowTransactionCommand(tx_hash.to_array(), neo_express_data_file=neoxp_path)
    with _NEOXP_CONFIG_LOCK:
        if isinstance(check_point_file, str):
            reset_neo_express_instance(neoxp_path, check_point_file)
        stdout, stderr = run_neo_express_cli(command)

    return stdout


def get_transaction(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None) -> Transaction:
    raw_result = _get_transaction_raw(neoxp_path, tx_hash, check_point_file)

    tx: Transaction
    try:
        import json
        result_json = json.loads(raw_result)
        tx = Transaction.from_json(result_json['transaction'])
    except:
        tx = None

    return tx


def get_transaction_log(neoxp_path: str, tx_hash: UInt256, check_point_file: str = None, contract_collection=None) -> TransactionLog:
    raw_result = _get_transaction_raw(neoxp_path, tx_hash, check_point_file)

    tx_log: TransactionLog
    try:
        import json
        result_json = json.loads(raw_result)
        tx_log = TransactionLog.from_json(result_json['application-log'], contract_collection)
        tx_log._tx_id = tx_hash
    except:
        tx_log = None

    return tx_log


def get_latest_block(neoxp_path: str, check_point_file: str = None) -> Block:
    return get_block(neoxp_path, check_point_file=check_point_file)


def get_block(neoxp_path: str, block_hash_or_index: Union[UInt256, int] = None,
              check_point_file: str = None) -> Block:
    if isinstance(block_hash_or_index, (UInt256, int)):
        if isinstance(block_hash_or_index, int):
            if block_hash_or_index < 0:
                block_hash_or_index = 0
        else:
            block_hash_or_index = block_hash_or_index.to_array()

        command = neoxp.show.ShowBlockCommand(block_hash_or_index,
                                              neo_express_data_file=neoxp_path)
    else:
        command = neoxp.show.ShowBlockCommand(neo_express_data_file=neoxp_path)

    with _NEOXP_CONFIG_LOCK:
        if isinstance(check_point_file, str):
            reset_neo_express_instance(neoxp_path, check_point_file)
        stdout, stderr = run_neo_express_cli(command)

    block: Block
    try:
        import json
        result_json = json.loads(stdout)
        block = Block.from_json(result_json)
    except:
        block = None

    return block


def run_batch(neoxp_path: str, batch_path: str, reset: bool = False, check_point_file: str = None) -> str:
    if isinstance(check_point_file, str) and not os.path.isfile(check_point_file):
        check_point_file = None
    command = neoxp.batch.BatchCommand(batch_path,
                                       neo_express_data_file=neoxp_path,
                                       check_point_file=check_point_file,
                                       reset=reset)
    with _NEOXP_CONFIG_LOCK:
        stdout, stderr = run_neo_express_cli(command)

    return stdout


def run_neo_express_cli(command: neoxp.NeoExpressCommand) -> Tuple[str, str]:
    neoxp_args = ['neoxp']
    neoxp_args.extend(command.cli_command().split())

    with _NEOXP_FILE_LOCK:
        process = subprocess.Popen(neoxp_args,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   text=True)
        cli_result = process.communicate()
    return cli_result



from __future__ import annotations

from abc import ABC

from boa3.internal.model.identifiedsymbol import IdentifiedSymbol


class IBuiltinSymbol(IdentifiedSymbol, ABC):

    def build(self, *args, **kwargs) -> IBuiltinSymbol:
        """
        Creates a symbol instance with the given value as self

        :return: The built method if the value is valid. The current object otherwise
        :rtype: IBuiltinSymbol
        """
        return self

    def update_with_analyser(self, analyser):
        return



from typing import Dict, Optional

from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand

__all__ = ['ResetCommand']


class ResetCommand(NeoExpressCommand):
    def __init__(self, node_index: Optional[int] = None,
                 force: bool = False,
                 reset_all: bool = False,
                 neo_express_data_file: str = None):

        self.node_index = node_index
        self.force = force
        self.reset_all = reset_all
        self.input = neo_express_data_file

        command_id = 'reset'
        args = []
        if isinstance(node_index, int):
            args.append(str(node_index))

        super().__init__(command_id, args)

    def _get_options(self) -> Dict[str, str]:
        options = super()._get_options()

        if self.force:
            options['--force'] = ''
        if self.reset_all:
            options['--all'] = ''

        return options



from __future__ import annotations

import ast
from typing import Dict, List, Optional

from boa3.builtin.compile_time import NeoMetadata
from boa3.internal import constants
from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.analyser.astoptimizer import AstOptimizer
from boa3.internal.analyser.constructanalyser import ConstructAnalyser
from boa3.internal.analyser.moduleanalyser import ModuleAnalyser
from boa3.internal.analyser.supportedstandard.standardanalyser import StandardAnalyser
from boa3.internal.analyser.typeanalyser import TypeAnalyser
from boa3.internal.compiler.compiledmetadata import CompiledMetadata
from boa3.internal.exception.CompilerError import CompilerError
from boa3.internal.exception.CompilerWarning import CompilerWarning
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.type import Type


class Analyser:
    """
    This class is responsible for the semantic analysis of the code

    :ivar symbol_table: a dictionary used to store the identifiers
    """

    def __init__(self, ast_tree: ast.AST, path: str = None, project_root: str = None,
                 env: str = None, log: bool = False, fail_fast: bool = False):
        self.symbol_table: Dict[str, ISymbol] = {}

        self.ast_tree: ast.AST = ast_tree
        self.metadata: NeoMetadata = NeoMetadata()
        self.is_analysed: bool = False
        self._log: bool = log
        self._fail_fast: bool = fail_fast
        self._env: str = env if env is not None else constants.DEFAULT_CONTRACT_ENVIRONMENT

        self.__include_builtins_symbols()
        self._errors = []
        self._warnings = []
        self._imported_files: Dict[str, Analyser] = {}
        self._included_imported_files: bool = False

        import os
        self.path: str = path
        self.filename: str = path if path is None else os.path.realpath(path)

        if project_root is not None:
            if not os.path.exists(project_root):
                project_root = os.path.abspath(f'{os.path.curdir}{os.path.sep}{project_root}')

            if os.path.isfile(project_root):
                project_root = os.path.dirname(os.path.abspath(project_root))

        self.root: str = (os.path.realpath(project_root)
                          if project_root is not None and os.path.isdir(project_root)
                          else path)

    @staticmethod
    def analyse(path: str, log: bool = False, fail_fast: bool = False,
                imported_files: Optional[Dict[str, Analyser]] = None,
                import_stack: Optional[List[str]] = None,
                root: str = None, env: str = None, compiler_entry: bool = False) -> Analyser:
        """
        Analyses the syntax of the Python code

        :param path: the path of the Python file
        :param log: if compiler errors should be logged.
        :param fail_fast: if should stop compilation on first error found.
        :param import_stack: a list that represents the current import stack if it's from an import.
                             If it's not triggered by an import, must be None.
        :param imported_files: a dict that maps the paths of the files that were analysed if it's from an import.
                               If it's not triggered by an import, must be None.
        :param root: the path of the project root that the current smart contract is part of.
        :param env: specific environment id to compile.
        :param compiler_entry: Whether this is the entry compiler analyser. False by default.
        :return: a boolean value that represents if the analysis was successful
        :rtype: Analyser
        """
        with open(path, 'rb') as source:
            ast_tree = ast.parse(source.read())

        analyser = Analyser(ast_tree, path, root if isinstance(root, str) else path, env, log, fail_fast)
        CompiledMetadata.set_current_metadata(analyser.metadata)

        if compiler_entry:
            from boa3.internal.model.imports.builtin import CompilerBuiltin
            CompilerBuiltin.update_with_analyser(analyser)

        analyser.__pre_execute()

        # fill symbol table
        if not analyser.__analyse_modules(imported_files, import_stack):
            return analyser
        analyser.__pre_execute()

        # check if standards are correctly implemented
        if not analyser.__check_standards():
            return analyser
        # check is the types are correct
        if not analyser.__check_types():
            return analyser

        analyser.__pos_execute()
        analyser.is_analysed = True

        return analyser

    @property
    def errors(self) -> List[CompilerError]:
        return self._errors.copy()

    @property
    def warnings(self) -> List[CompilerWarning]:
        return self._warnings.copy()

    @property
    def env(self) -> str:
        return self._env

    def copy(self) -> Analyser:
        copied = Analyser(ast_tree=self.ast_tree, path=self.path, project_root=self.root,
                          env=self._env, log=self._log, fail_fast=self._fail_fast)

        copied.metadata = self.metadata
        copied.is_analysed = self.is_analysed
        copied.symbol_table = self.symbol_table.copy()
        copied.filename = self.filename
        copied._imported_files = self._imported_files.copy()
        copied._included_imported_files = self._included_imported_files

        return copied

    def __include_builtins_symbols(self):
        """
        Include the Python builtins in the global symbol table
        """
        self.symbol_table.update(Type.builtin_types())

    def __check_types(self) -> bool:
        """
        Performs the type checking

        :return: a boolean value that represents if the analysis was successful
        """
        type_analyser = TypeAnalyser(self, self.symbol_table, log=self._log, fail_fast=self._fail_fast)
        self._update_logs(type_analyser)
        return not type_analyser.has_errors

    def __analyse_modules(self,
                          imported_files: Optional[Dict[str, Analyser]] = None,
                          import_stack: Optional[List[str]] = None) -> bool:
        """
        Validates the symbols and constructs the symbol table of the ast tree

        :return: a boolean value that represents if the analysis was successful
        """
        current_metadata = self.metadata
        module_analyser = ModuleAnalyser(self, self.symbol_table,
                                         log=self._log,
                                         fail_fast=self._fail_fast,
                                         filename=self.filename,
                                         root_folder=self.root,
                                         analysed_files=imported_files,
                                         import_stack=import_stack)
        self.symbol_table.update(module_analyser.global_symbols)
        self.ast_tree.body.extend(module_analyser.imported_nodes)
        self._update_logs(module_analyser)
        self._imported_files = module_analyser.analysed_files.copy()

        if self.metadata != current_metadata:
            CompiledMetadata.set_current_metadata(self.metadata)

        return not module_analyser.has_errors

    def __check_standards(self) -> bool:
        """
        Verify if the standards included in the metadata are fully implemented

        :return: a boolean value that represents if the analysis was successful
        """
        standards_analyser = StandardAnalyser(self, self.symbol_table, log=self._log, fail_fast=self._fail_fast)
        self._update_logs(standards_analyser)
        return not standards_analyser.has_errors

    def _update_logs(self, analyser: IAstAnalyser):
        self._errors.extend(analyser.errors)
        self._warnings.extend(analyser.warnings)

    def __pre_execute(self):
        """
        Pre executes the instructions of the ast for optimization
        """
        self.ast_tree = ConstructAnalyser(self, self.ast_tree, self.symbol_table,
                                          log=self._log, fail_fast=self._fail_fast
                                          ).tree

    def __pos_execute(self):
        """
        Tries to optimize the ast after validations
        """
        optimizer = AstOptimizer(self, log=self._log, fail_fast=self._fail_fast)
        self._update_logs(optimizer)

    def update_symbol_table(self, symbol_table: Dict[str, ISymbol]):
        for symbol_id, symbol in symbol_table.items():
            if (hasattr(symbol, 'origin')
                    and hasattr(symbol.origin, 'origin')
                    and isinstance(symbol.origin.origin, ast.AST)
                    and len(symbol_id.split(constants.VARIABLE_NAME_SEPARATOR)) <= 1):

                if symbol_id in self.symbol_table:
                    self.symbol_table.pop(symbol_id)

                origin_hash = symbol.origin.origin.__hash__()
                unique_id = '{0}{2}{1}'.format(origin_hash, symbol_id, constants.VARIABLE_NAME_SEPARATOR)
            else:
                unique_id = symbol_id

            from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
            from boa3.internal.model.type.classes.userclass import UserClass
            if not isinstance(symbol, IdentifiedSymbol) or isinstance(symbol, UserClass):
                if unique_id not in self.symbol_table:
                    self.symbol_table[unique_id] = symbol

    def update_symbol_table_with_imports(self):
        if self._included_imported_files:
            return

        import os
        from boa3.internal.analyser.importanalyser import ImportAnalyser
        from boa3.internal.model.imports.importsymbol import Import

        imports = self._imported_files.copy()
        paths_already_imported = [imported.origin for imported in self.symbol_table.values()
                                  if isinstance(imported, Import) and isinstance(imported.origin, str)]
        if isinstance(self.path, str):
            paths_already_imported.append(self.path.replace(os.path.sep, constants.PATH_SEPARATOR))

        for file_path, analyser in imports.items():
            if file_path not in paths_already_imported:
                import_analyser = ImportAnalyser(file_path, self.root,
                                                 already_imported_modules=imports,
                                                 log=False, fail_fast=self._fail_fast,
                                                 get_entry=True)
                import_symbol = Import(file_path, analyser.ast_tree, import_analyser, {})
                self.symbol_table[file_path] = import_symbol

        self._included_imported_files = True

    def get_imports(self) -> List[Analyser]:
        return list(self._imported_files.values())



from typing import Dict, List, Optional, Tuple

from boa3.internal.model.builtin.classmethod.countmethod import CountMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode import OpcodeHelper
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class CountSequenceMethod(CountMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(sequence_type, SequenceType):
            sequence_type = Type.sequence
        if not isinstance(arg_value, IType):
            arg_value = Type.any

        args: Dict[str, Variable] = {
            'self': Variable(sequence_type),
            'value': Variable(arg_value)
        }

        self._generic_verification_opcodes = None

        super().__init__(args)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        sequence_type: IType = params[0].type

        if not isinstance(sequence_type, SequenceType):
            return False
        from boa3.internal.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.internal.model.type.collection.sequence.tupletype import TupleType
        from boa3.internal.model.type.collection.sequence.rangetype import RangeType
        if not isinstance(sequence_type, (ListType, TupleType, RangeType)):
            return False
        return True

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')

        # region Sequence logic

        repack_array = [  # recreates an array to not change the original one
            (Opcode.UNPACK, b''),
            (Opcode.PACK, b''),
        ]

        sequence_initialize = [  # initializes variables
            (Opcode.PUSH0, b''),  # count = 0
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),  # index = len(sequence) - -1
        ]

        sequence_verify_while = [  # verifies if all the elements from the sequence were visited
            (Opcode.DUP, b''),  # would be equivalent to: while (index >= 0)
            (Opcode.SIGN, b''),
            (Opcode.PUSH0, b''),  # if index < 0:
            jmp_place_holder,  # jump to clean the stack
        ]

        sequence_get_element = [  # gets a element from the sequence
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b'')  # element = sequence[index]
        ]

        sequence_equals = [  # verifies if the element and the given value are the same
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.EQUAL, b''),  # if element != value:
            jmp_place_holder  # jump to list_tuple_count_index_dec
        ]

        sequence_count_inc = [  # increment count if element == value
            (Opcode.SWAP, b''),
            (Opcode.INC, b''),  # count++
            (Opcode.SWAP, b''),
        ]

        num_jmp_code = get_bytes_count(sequence_count_inc)
        jmp_to_dec_statement = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
        sequence_equals[-1] = jmp_to_dec_statement

        list_tuple_count_index_dec = [  # decreases the index
            (Opcode.DEC, b''),  # index--
            # return to the while verification
        ]

        in_depth_verification = self.generic_verification(get_bytes_count(sequence_count_inc),
                                                          get_bytes_count(sequence_get_element + sequence_equals))

        num_jmp_code = -get_bytes_count(list_tuple_count_index_dec + sequence_count_inc + sequence_equals +
                                        sequence_get_element + sequence_verify_while + in_depth_verification)
        jmp_back_to_while_verify_statement = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code)
        list_tuple_count_index_dec.append(jmp_back_to_while_verify_statement)

        num_jmp_code = get_bytes_count(sequence_get_element + sequence_equals +
                                       sequence_count_inc + list_tuple_count_index_dec + in_depth_verification)
        jmp_to_clean_statement = OpcodeHelper.get_jump_and_data(Opcode.JMPLT, num_jmp_code, True)
        sequence_verify_while[-1] = jmp_to_clean_statement

        sequence_clean_stack = [
            (Opcode.DROP, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        # endregion

        return (
            repack_array +
            sequence_initialize +
            sequence_verify_while +
            in_depth_verification +
            sequence_get_element +
            sequence_equals +
            sequence_count_inc +
            list_tuple_count_index_dec +
            sequence_clean_stack
        )

    def generic_verification(self, inc_statement_bytes=None,
                             get_equals_statement_bytes=None) -> List[Tuple[Opcode, bytes]]:
        if self._generic_verification_opcodes is None:
            self._generic_verification_opcodes = []

        return self._generic_verification_opcodes



from typing import Any

from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.strtype import StrType


class AddressType(StrType):
    """
    A class used to indicate that a parameter or return on the manifest is an Address. It's a subclass of StrType.
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'Address'

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _Address


_Address = AddressType()



from __future__ import annotations

from typing import Any, List, Optional, Tuple

from boa3.internal.model.builtin.interop.interopinterfacetype import InteropInterfaceType
from boa3.internal.model.method import Method
from boa3.internal.model.type.collection.icollection import ICollectionType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IteratorType(InteropInterfaceType, ICollectionType):
    """
    A class used to represent Neo Iterator class
    """

    def __init__(self, collection: ICollectionType = None):
        super().__init__('Iterator')

        if collection is None:
            from boa3.internal.model.type.type import Type
            collection = Type.collection

        self._constructor: Method = None
        self._origin_collection: ICollectionType = collection

        self._methods = None
        self._properties = None

        self.key_type = self._origin_collection.valid_key
        self.item_type = self._origin_collection.item_type

    @property
    def identifier(self) -> str:
        return '{0}[{1}, {2}]'.format(self._identifier, self.valid_key.identifier, self.item_type.identifier)

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @property
    def value_type(self) -> IType:
        return self._origin_collection.item_type

    @property
    def valid_key(self) -> IType:
        return self._origin_collection.valid_key

    def is_valid_key(self, key_type: IType) -> bool:
        return self._origin_collection.is_valid_key(key_type)

    @classmethod
    def build(cls, value: Any = None) -> IteratorType:
        if isinstance(value, ICollectionType):
            return IteratorType(value)
        else:
            return _Iterator

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, IteratorType)

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.internal.model.builtin.interop.iterator.iteratornextmethod import IteratorNextMethod
        from boa3.internal.model.builtin.interop.iterator.getiteratorvalue import IteratorValueProperty

        self._instance_methods['next'] = IteratorNextMethod()

        self._properties['value'] = IteratorValueProperty(self)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.model.type.classes.pythonclass import PythonClass
        return super(PythonClass, self).is_instance_opcodes()

    def __eq__(self, other) -> bool:
        if not isinstance(other, IteratorType):
            return False
        return self.valid_key == other.valid_key and self.value_type == other.value_type

    def __hash__(self) -> int:
        return self._origin_collection.__hash__()


_Iterator = IteratorType()



from typing import Any, Dict, List, Optional, Sized, Tuple

from boa3.internal.compiler.codegenerator import get_bytes_count
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.model.type.type import IType, Type
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode import OpcodeHelper
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ScriptHashMethod(IBuiltinMethod):

    def __init__(self, data_type: IType = None):
        if (not Type.int.is_type_of(data_type)
                and not Type.str.is_type_of(data_type)
                and not Type.bytes.is_type_of(data_type)):
            data_type = Type.any

        identifier = 'to_script_hash'
        args: Dict[str, Variable] = {'value': Variable(data_type)}
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        super().__init__(identifier, args, return_type=UInt160Type.build())

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        self_type = self.args['value'].type
        if self_type is Type.any:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self_type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, PrimitiveType)

    @property
    def is_supported(self) -> bool:
        return self.args['value'].type is not Type.any

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.constants import SIZE_OF_INT160
        from boa3.internal.model.builtin.interop.crypto import Sha256Method
        from boa3.internal.model.builtin.interop.crypto import Ripemd160Method
        from boa3.internal.model.builtin.interop.stdlib.base58decodemethod import Base58DecodeMethod
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.Integer import Integer

        raise_exception_if_size_is_wrong = [
            (Opcode.JMP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.THROW, b''),
        ]
        raise_exception_if_size_is_wrong[0] = OpcodeHelper.get_jump_and_data(Opcode.JMPGE, get_bytes_count(raise_exception_if_size_is_wrong))

        except_use_hash_160 = Sha256Method().opcode + Ripemd160Method().opcode
        end_try = [
            (Opcode.ENDTRY, b'')
        ]
        end_try[-1] = OpcodeHelper.get_jump_and_data(Opcode.ENDTRY, get_bytes_count(end_try))

        try_to_decode_with_base58 = [
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, Type.str.stack_item),
            (Opcode.JMPIF, Integer(4).to_byte_array(min_length=1)),
            (Opcode.CONVERT, Type.str.stack_item),
            (Opcode.DUP, b''),
        ] + Base58DecodeMethod().opcode + [
            (Opcode.DUP, b''),  # check if size is UInt160 + 1  // first position will be ignored
            (Opcode.SIZE, b''),
            OpcodeHelper.get_push_and_data(SIZE_OF_INT160 + 1),
            (Opcode.SWAP, b''),
            (Opcode.OVER, b''),
        ] + raise_exception_if_size_is_wrong + [
            (Opcode.DEC, b''),
            OpcodeHelper.get_push_and_data(1),
            (Opcode.SWAP, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, Type.str.stack_item),
            (Opcode.NIP, b''),
            OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(except_use_hash_160)),
        ]

        try_code = OpcodeHelper.get_try_and_data(get_bytes_count(try_to_decode_with_base58), jump_through=True)
        begin_try = (
            [
                try_code
            ]
            + try_to_decode_with_base58
            + except_use_hash_160
            + end_try
        )

        return begin_try

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if 'value' in self.args and self.args['value'].type is not Type.any:
            return self

        from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
        from boa3.internal.model.type.primitive.inttype import IntType
        from boa3.internal.model.type.primitive.strtype import StrType
        from boa3.internal.model.type.primitive.bytestype import BytesType

        if isinstance(value, Sized) and len(value) == 1:
            value = value[0]

        if isinstance(value, ECPointType):
            from boa3.internal.model.builtin.method.ecpointtoscripthashmethod import ECPointToScriptHashMethod
            return ECPointToScriptHashMethod()
        elif isinstance(value, (IntType, StrType, BytesType)):
            return ScriptHashMethod(value)
        elif isinstance(value, IType):
            return ScriptHashMethod(Type.bytes)
        return super().build(value)



from typing import List, Tuple

from boa3.internal.model.builtin.method.printmethod import PrintMethod
from boa3.internal.model.type.classes.userclass import UserClass
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.opcode import OpcodeHelper
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintClassMethod(PrintMethod):

    def __init__(self, arg_value: IType):
        if not isinstance(arg_value, UserClass):
            from boa3.internal.model.type.classes import userclass
            arg_value = userclass._EMPTY_CLASS

        super().__init__(arg_value)
        self._arg_type = arg_value

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:

            class_to_dict = [
                (Opcode.UNPACK, b''),
                (Opcode.DROP, b''),
                (Opcode.NEWMAP, b'')
            ]
            for variable_name in self._arg_type.instance_variables:
                class_to_dict.extend([
                    (Opcode.TUCK, b''),
                    OpcodeHelper.get_pushdata_and_data(variable_name),
                    (Opcode.ROT, b''),
                    (Opcode.SETITEM, b'')
                ])

            from boa3.internal.model.builtin.interop.interop import Interop
            self._print_value_opcodes = (
                class_to_dict
                + Interop.JsonSerialize.opcode
            )

        return super().print_value_opcodes



import json
from typing import Any, Dict

from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.icontractcommand import IContractCommand


class ContractDeployCommand(IContractCommand):
    def __init__(self, contract_nef: str, account: Account,
                 witness_scope: WitnessScope = WitnessScope.CalledByEntry,
                 data: Any = None,
                 wallet_password: str = None,
                 trace: bool = False,
                 force: bool = False,
                 json_output: bool = False,
                 neo_express_data_file: str = None):

        self.contract_nef = contract_nef
        self.account = account

        self.witness_scope = witness_scope
        self.data = data
        self.password = wallet_password

        self.input = neo_express_data_file
        self.trace = trace
        self.force = force
        self.json_output = json_output

        super().__init__('deploy', [contract_nef, account.get_identifier()])

    def _get_options(self) -> Dict[str, str]:
        options = super()._get_options()

        if isinstance(self.witness_scope, WitnessScope):
            options['--witness-scope'] = self.witness_scope.name
        if self.data is not None:
            options['--data'] = json.dumps(self.data)
        if isinstance(self.password, str):
            options['--password'] = self.password
        if self.trace:
            options['--trace'] = ''
        if self.force:
            options['--force'] = ''
        if self.json_output:
            options['--json'] = ''

        return options



import boa3.builtin.compile_time
import boa3.builtin.contract
import boa3.builtin.interop
import boa3.builtin.math
import boa3.builtin.nativecontract
import boa3.builtin.vm


env: str
"""
Gets the compiled environment. This allows specific environment validations to be easily included in the smart contract 
logic without the need to rewrite anything before compiling (i.e. changes in smart contracts hashes between testnet and 
mainnet). 

>>> # compiling with 'neo3-boa compile -e test_net ./path/to/contract.py'
... from boa3.builtin.interop.contract import call_contract
... from boa3.builtin.type import UInt160
... call_contract(UInt160(b'12345678901234567890') if env == 'test_net' else b'abcdeabcdeabcdeabcde',
...               'balanceOf',
...               UInt160(b'zyxwvzyxwvzyxwvzyxwv'))
110000

>>> # compiling with 'neo3-boa compile -e main_net ./path/to/contract.py'
... from boa3.builtin.interop.contract import call_contract
... from boa3.builtin.type import UInt160
... call_contract(UInt160(b'12345678901234567890') if env == 'test_net' else b'abcdeabcdeabcdeabcde',
...               'balanceOf',
...               UInt160(b'zyxwvzyxwvzyxwvzyxwv'))
250

:meta hide-value:
"""



import json
import os.path
import subprocess
import threading
from typing import Any, List, Tuple, Dict, Optional, Union, Type

from boa3.internal import env, constants
from boa3.internal.neo import utils as neo_utils
from boa3.internal.neo.vm.type.String import String
from boa3.internal.neo3.core.types import UInt256
from boa3.internal.neo3.vm import vmstate, VMState
from boa3_test.test_drive.model.invoker.neobatchinvoke import NeoBatchInvoke
from boa3_test.test_drive.model.invoker.neoinvokecollection import NeoInvokeCollection
from boa3_test.test_drive.model.invoker.neoinvokeresult import NeoInvokeResult
from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope
from boa3_test.test_drive.model.smart_contract.contractcollection import ContractCollection
from boa3_test.test_drive.model.smart_contract.testcontract import TestContract
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp import utils as neoxp_utils
from boa3_test.test_drive.neoxp.batch import NeoExpressBatch
from boa3_test.test_drive.testrunner import utils
from boa3_test.test_drive.testrunner.blockchain.block import TestRunnerBlock as Block
from boa3_test.test_drive.testrunner.blockchain.log import TestRunnerLog as Log
from boa3_test.test_drive.testrunner.blockchain.notification import TestRunnerNotification as Notification
from boa3_test.test_drive.testrunner.blockchain.storage import TestRunnerStorage as Storage
from boa3_test.test_drive.testrunner.blockchain.storagecollection import StorageCollection
from boa3_test.test_drive.testrunner.blockchain.transaction import TestRunnerTransaction as Transaction
from boa3_test.test_drive.testrunner.blockchain.transactionlog import TestRunnerTransactionLog as TransactionLog


class NeoTestRunner:
    _FOLDER_NAME = 'test-runner'
    _ROOT_FOLDER = os.path.abspath(os.path.curdir) + os.path.sep + _FOLDER_NAME
    _INVOKE_FILE = f'{_FOLDER_NAME}.neo-invoke.json'
    _BATCH_FILE = f'{_FOLDER_NAME}.batch'
    _CHECKPOINT_FILE = f'{_FOLDER_NAME}.neoxp-checkpoint'

    _DEFAULT_ACCOUNT = neoxp_utils.get_default_account()

    def __init__(self, neoxp_path: str = None, runner_id: str = None):
        self._vm_state: VMState = VMState.NONE
        self._gas_consumed: int = 0
        self._result_stack: List[Any] = []
        self._error_message: Optional[str] = None
        self._last_cli_log: Optional[str] = None
        self._cli_log: str = ''

        self._calling_account: Optional[Account] = None
        self._notifications: List[Notification] = []
        self._logs: List[Log] = []
        self._storages: StorageCollection = StorageCollection()

        if not isinstance(neoxp_path, str):
            neoxp_path = f'{env.NEO_EXPRESS_INSTANCE_DIRECTORY}{os.path.sep}default.neo-express'
        self._neoxp_abs_path = os.path.abspath(neoxp_path)

        if isinstance(runner_id, str):
            self._file_name: str = None  # defined in the following line
            self.file_name = self._FOLDER_NAME if not runner_id else runner_id
        else:
            self._file_name = self._FOLDER_NAME

        self._clear_files_when_destroyed = True
        self._first_execution = True

        self._batch = NeoExpressBatch()
        self._contracts = ContractCollection()
        self._invokes = NeoInvokeCollection()
        self._invokes_to_batch = 0
        self._last_execution_results: List[NeoInvokeResult] = []

    @property
    def file_name(self) -> str:
        return self._file_name

    @file_name.setter
    def file_name(self, value: str):
        if isinstance(value, str) and not value.isspace():
            from boa3_test.test_drive import utils
            runner_specific_id = utils.create_custom_id(value)
            self._file_name = value
            self._INVOKE_FILE = f'{runner_specific_id}.neo-invoke.json'
            self._BATCH_FILE = f'{runner_specific_id}.batch'
            self._CHECKPOINT_FILE = f'{runner_specific_id}.neoxp-checkpoint'

    @property
    def vm_state(self) -> VMState:
        return self._vm_state

    @property
    def gas_consumed(self) -> int:
        return self._gas_consumed

    @property
    def result_stack(self) -> List[Any]:
        return self._result_stack.copy()

    @property
    def error(self) -> Optional[str]:
        return self._error_message

    @property
    def cli_log(self) -> str:
        return self._cli_log

    @property
    def notifications(self) -> List[Notification]:
        return self._notifications.copy()

    def get_events(self, event_name: str = None, origin: TestContract = None) -> List[Notification]:
        return self._filter_events(self._notifications, event_name, origin)

    def get_logs(self, origin: TestContract = None) -> List[Log]:
        return self._filter_events(self._logs, origin=origin)

    def _filter_events(self, events: list, event_name: str = None, origin: TestContract = None) -> list:
        if origin is None and event_name is None:
            return events.copy()
        elif origin is None:
            return [n for n in events if n.name == event_name]
        else:
            if hasattr(origin, 'script_hash'):
                origin_bytes = origin.script_hash
            elif hasattr(origin, 'to_array'):
                origin_bytes = origin.to_array()
            else:
                origin_bytes = bytes(origin)

            if event_name is None:
                return [n for n in events if n.origin == origin_bytes]
            else:
                return [n for n in events if (n.name == event_name
                                              and n.origin == origin_bytes)]

    @property
    def logs(self) -> List[Log]:
        return self._logs.copy()

    @property
    def contracts(self) -> ContractCollection:
        return self._contracts

    @property
    def storages(self) -> StorageCollection:
        return self._storages

    @property
    def _root(self) -> str:
        return self._ROOT_FOLDER

    @_root.setter
    def _root(self, value: str):
        if isinstance(value, str) and os.path.exists(value) and os.path.isdir(value):
            sep = os.path.sep
            root = os.path.abspath(value) + sep + self._FOLDER_NAME
            self._ROOT_FOLDER = root

    def _update_cli_log(self, log_to_append: str):
        if self._last_cli_log is None:
            log_connector = '\n' if len(self._cli_log) > 0 and len(log_to_append) > 0 else ''
            self._cli_log += f'{log_connector}{log_to_append}'
        else:
            self._last_cli_log = self._cli_log
            self._cli_log = log_to_append

    def add_neo(self, script_hash_or_address: Union[bytes, str], amount: int):
        address = neoxp_utils.get_account_from_script_hash_or_id(script_hash_or_address)
        self._batch.transfer_assets(sender=self._DEFAULT_ACCOUNT, receiver=address,
                                    quantity=amount,
                                    asset='NEO')

    def add_gas(self, script_hash_or_address: Union[bytes, str], amount: int):
        address = neoxp_utils.get_account_from_script_hash_or_id(script_hash_or_address)
        gas_decimals = 8
        self._batch.transfer_assets(sender=self._DEFAULT_ACCOUNT, receiver=address,
                                    asset='GAS', decimals=gas_decimals,
                                    quantity=(amount / (10 ** gas_decimals)))

    def get_genesis_block(self) -> Block:
        return self.get_block(0)

    def get_latest_block(self) -> Block:
        return self.get_block(None)

    def get_block(self, block_hash_or_index: Union[UInt256, bytes, int]) -> Optional[Block]:
        genesis = neoxp_utils.get_genesis_block()
        if isinstance(genesis, Block) and block_hash_or_index in (genesis.hash, genesis.index):
            # genesis block doesn't change between neo express resets
            return genesis

        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        block = neoxp_utils.get_block(self._neoxp_abs_path, block_hash_or_index,
                                      check_point_file=check_point_path)

        if not isinstance(genesis, Block) and isinstance(block, Block) and block.index == 0:
            neoxp_utils._set_genesis_block(block)  # optimization for consecutive executions
        return block

    def get_transaction(self, tx_hash: Union[UInt256, bytes]) -> Optional[Transaction]:
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        return neoxp_utils.get_transaction(self._neoxp_abs_path, tx_hash,
                                           check_point_file=check_point_path)

    def get_transaction_result(self, tx_hash: Union[UInt256, bytes]) -> Optional[TransactionLog]:
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)
        return neoxp_utils.get_transaction_log(self._neoxp_abs_path, tx_hash,
                                               check_point_file=check_point_path,
                                               contract_collection=self._contracts)

    def deploy_contract(self, nef_path: str, account: Account = None) -> TestContract:
        if not isinstance(nef_path, str) or not nef_path.endswith('.nef'):
            raise ValueError('Requires a .nef file to deploy a contract')

        if nef_path not in self._contracts:
            contract = self._batch.deploy_contract(nef_path, account)
            if contract.name in self._contracts:
                raise ValueError('Contract with duplicated name')
            self._contracts.append(contract)
        else:
            contract = self._contracts[nef_path]
        return contract

    def call_contract(self, nef_path: str, method: str, *arguments: Any,
                      expected_result_type: Type = None) -> NeoInvokeResult:
        if nef_path not in self._contracts:
            contract = self.deploy_contract(nef_path)
        else:
            contract = self._contracts[nef_path]

        return self._invokes.append_contract_invoke(contract, method, *arguments,
                                                    expected_result_type=expected_result_type)

    def run_contract(self, nef_path: str, method: str, *arguments: Any,
                     account: Account = None, witness_scope: WitnessScope = WitnessScope.CalledByEntry,
                     expected_result_type: Type = None) -> NeoBatchInvoke:
        if nef_path not in self._contracts:
            contract = self.deploy_contract(nef_path)
        else:
            contract = self._contracts[nef_path]

        if witness_scope != WitnessScope.CalledByEntry:
            # neo express only supports CalledByEntry and Global as witness scopes
            witness_scope = WitnessScope.Global

        invoke = self._invokes.create_contract_invoke(contract, method, *arguments)
        if isinstance(account, Account):
            invoke._invoker = account
        return self._batch.run_contract(invoke,
                                        witness_scope=witness_scope,
                                        expected_result_type=expected_result_type)

    def get_contract(self, contract_id: Union[str, bytes]) -> TestContract:
        return self._contracts[contract_id]

    def update_contracts(self, export_checkpoint: bool = False):
        self._generate_root_folder()
        if export_checkpoint:
            self._create_checkpoint_from_batch()
        else:
            self._update_contracts()

    def _update_contracts(self):
        batch_file_path = self.get_full_path(self._BATCH_FILE)
        if self._first_execution:
            check_point_path = None
            self._first_execution = False
        else:
            check_point_path = self.get_full_path(self._CHECKPOINT_FILE)

        batch_has_deploys = self._batch.has_new_deploys()
        log = self._batch.execute(self._neoxp_abs_path, batch_file_path,
                                  check_point_file=check_point_path,
                                  reset=True
                                  )
        self._update_cli_log(log)

        if batch_has_deploys:
            self._contracts.update_after_deploy()

    def execute(self, account: Account = None, get_storage_from: Union[str, TestContract] = None,
                clear_invokes: bool = True, add_invokes_to_batch: bool = False):
        self._generate_files()
        invoke_file_path = self.get_full_path(self._INVOKE_FILE)
        cli_args = ['neo-test-runner', invoke_file_path
                    ]

        checkpoint_file = self.get_full_path(self._CHECKPOINT_FILE)
        if os.path.isfile(checkpoint_file):
            cli_args.extend(['--checkpoint', checkpoint_file])

        if isinstance(account, Account):
            cli_args.extend(['--account', account.get_identifier()])
            cli_args.extend(['--express', self._neoxp_abs_path])
        elif account is not None:
            account = None
        self._calling_account = account

        if isinstance(get_storage_from, TestContract):
            get_storage_from = get_storage_from.get_identifier()
        if isinstance(get_storage_from, str):
            cli_args.extend(['--storages', get_storage_from])
        stdout, stderr = self._run_command_line(cli_args)

        try:
            self.reset_state()
            try:
                result = json.loads(stdout)
            except json.JSONDecodeError:
                result = utils.handle_return_error(stdout)

            self._update_runner(result)
        except BaseException:
            self._error_message = stdout

        if add_invokes_to_batch:
            import shutil
            self._invokes_to_batch += 1
            invoke_file_path_to_batch = self.get_full_path(f'{self._file_name}_{self._invokes_to_batch}.neo-invoke.json')
            shutil.copy(invoke_file_path, invoke_file_path_to_batch)
            self._batch.invoke_file(invoke_file_path_to_batch, account=account)

        if clear_invokes:
            self._invokes.clear(state=self._vm_state)

        if self._error_message is not None:
            return self._error_message
        else:
            return self._last_execution_results

    def __del__(self):
        self.reset()
        if self._clear_files_when_destroyed:
            paths_to_delete = [
                self.get_full_path(self._CHECKPOINT_FILE),
                self.get_full_path(self._BATCH_FILE),
                self.get_full_path(self._INVOKE_FILE)
            ]
            for path in paths_to_delete:
                try:
                    if os.path.exists(path):
                        os.remove(path)
                except BaseException as e:
                    print(e)
                    continue

    def reset_state(self):
        self._vm_state = VMState.NONE
        self._gas_consumed = 0
        self._result_stack = []
        self._last_execution_results = []
        self._notifications.clear()
        self._logs.clear()
        self._storages.clear()
        self._last_cli_log = None
        self._error_message = None

    def reset(self):
        self.reset_state()
        self._first_execution = True

        self._batch.clear()
        self._contracts.clear()
        self._invokes.clear()
        self._invokes_to_batch = 0

    def _update_runner(self, result: Dict[str, Any]):
        self.reset_state()
        self._error_message = result['exception'] if 'exception' in result else None

        self._last_cli_log = self._cli_log
        if 'state' in result:
            self._vm_state = vmstate.get_vm_state(result['state'])

        if 'gasconsumed' in result:
            self._gas_consumed = int(float(result['gasconsumed']) * 10 ** constants.GAS_DECIMALS)

        if 'stack' in result:
            result_stack = result['stack']
            if isinstance(result_stack, list):
                new_result_stack = [neo_utils.stack_item_from_json(value) for value in result_stack]
            else:
                new_result_stack = [neo_utils.stack_item_from_json(result_stack)]

            self._last_execution_results = self._invokes.set_results(new_result_stack,
                                                                     calling_account=self._calling_account)
            self._result_stack = new_result_stack

        if 'notifications' in result:
            json_notifications = result['notifications']
            if not isinstance(json_notifications, list):
                json_notifications = [json_notifications]

            notifications = []
            for n in json_notifications:
                new = Notification.from_json(n, contract_collection=self._contracts)
                if new is not None:
                    notifications.append(new)
            self._notifications.extend(notifications)

        if 'logs' in result:
            json_logs = result['logs']
            if not isinstance(json_logs, list):
                json_logs = [json_logs]

            logs = []
            for l in json_logs:
                new = Log.from_json(l)
                if new is not None:
                    logs.append(new)
            self._logs.extend(logs)

        if 'storages' in result:
            json_storages = result['storages']
            if not isinstance(json_storages, list):
                json_storages = [json_storages]

            for s in json_storages:
                new = Storage.from_json(s, self._contracts)
                if new is not None:
                    self._storages.append(new)

    def _generate_root_folder(self):
        if not os.path.exists(self._root):
            os.mkdir(self._root)

    def _generate_files(self):
        self._generate_root_folder()

        checkpoint_thread = threading.Thread(target=self._create_checkpoint_from_batch)
        checkpoint_thread.start()

        self._generate_invoke_file()
        checkpoint_thread.join()

    def _create_checkpoint_from_batch(self):
        check_point_path = self.get_full_path(self._CHECKPOINT_FILE)

        if not os.path.exists(check_point_path) or self._batch.cur_size() > 0:
            self._batch.create_neo_express_checkpoint(check_point_path,
                                                      overwrite=True)
        # update list of contracts
        self._update_contracts()

    def _generate_invoke_file(self):
        invoke_file_content = self._invokes.to_json()
        param_json = json.dumps(invoke_file_content, separators=(',', ':'))
        with open(self.get_full_path(self._INVOKE_FILE), 'wb+') as json_file:
            json_file.write(String(param_json).to_bytes())

    def get_full_path(self, file_name: str):
        return self._root + os.path.sep + file_name

    def _run_command_line(self, args: List[str]) -> Tuple[str, str]:
        process = subprocess.Popen(args,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   text=True)
        return process.communicate()

    def increase_block(self, block_to_mint: int = None, time_interval_in_secs: int = 0):
        self._batch.mint_block(block_to_mint, time_interval_in_secs)



__all__ = [
    'NamedCurve',
    'sha256',
    'ripemd160',
    'hash160',
    'hash256',
    'check_sig',
    'check_multisig',
    'verify_with_ecdsa',
    'murmur32',
]


from typing import Any, List

from boa3.builtin.interop.crypto.namedcurve import NamedCurve
from boa3.builtin.type import ECPoint


def sha256(key: Any) -> bytes:
    """
    Encrypts a key using SHA-256.

    >>> sha256('unit test')
    b'\\xdau1>J\\xc2W\\xf8LN\\xfb2\\x0f\\xbd\\x01\\x1cr@<\\xf5\\x93<\\x90\\xd2\\xe3\\xb8$\\xd6H\\x96\\xf8\\x9a'

    >>> sha256(10)
    b'\\x9c\\x82r\\x01\\xb9@\\x19\\xb4/\\x85pk\\xc4\\x9cY\\xff\\x84\\xb5`M\\x11\\xca\\xaf\\xb9\\n\\xb9HV\\xc4\\xe1\\xddz'

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def ripemd160(key: Any) -> bytes:
    """
    Encrypts a key using RIPEMD-160.

    >>> ripemd160('unit test')
    b'H\\x8e\\xef\\xf4Zh\\x89:\\xe6\\xf1\\xdc\\x08\\xdd\\x8f\\x01\\rD\\n\\xbdH'

    >>> ripemd160(10)
    b'\\xc0\\xda\\x02P8\\xed\\x83\\xc6\\x87\\xdd\\xc40\\xda\\x98F\\xec\\xb9\\x7f9\\x98'

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def hash160(key: Any) -> bytes:
    """
    Encrypts a key using HASH160.

    >>> hash160('unit test')
    b'#Q\\xc9\\xaf+c\\x12\\xb1\\xb9\\x9e\\xa1\\x89t\\xa228g\\xec\\x0eF'

    >>> hash160(10)
    b'\\x89\\x86D\\x19\\xa8\\xc3v%\\x00\\xfe\\x9a\\x98\\xaf\\x8f\\xbbO3u\\x08\\xf0'

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def hash256(key: Any) -> bytes:
    """
    Encrypts a key using HASH256.

    >>> hash256('unit test')
    b'\\xdau1>J\\xc2W\\xf8LN\\xfb2\\x0f\\xbd\\x01\\x1cr@<\\xf5\\x93<\\x90\\xd2\\xe3\\xb8$\\xd6H\\x96\\xf8\\x9a'

    >>> hash256(10)
    b'\\x9c\\x82r\\x01\\xb9@\\x19\\xb4/\\x85pk\\xc4\\x9cY\\xff\\x84\\xb5`M\\x11\\xca\\xaf\\xb9\\n\\xb9HV\\xc4\\xe1\\xddz'

    :param key: the key to be encrypted
    :type key: Any
    :return: a byte value that represents the encrypted key
    :rtype: bytes
    """
    pass


def check_sig(pub_key: ECPoint, signature: bytes) -> bool:
    """
    Checks the signature for the current script container.

    >>> check_sig(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'),
    ...           b'wrongsignature')
    False

    :param pub_key: the public key of the account
    :type pub_key: ECPoint
    :param signature: the signature of the current script container
    :type signature: bytes
    :return: whether the signature is valid or not
    :rtype: bool
    """
    pass


def check_multisig(pubkeys: List[ECPoint], signatures: List[bytes]) -> bool:
    """
    Checks the signatures for the current script container.

    >>> check_multisig([ECPoint(b"\\x03\\xcd\\xb0\\x67\\xd9\\x30\\xfd\\x5a\\xda\\xa6\\xc6\\x85\\x45\\x01\\x60\\x44\\xaa\\xdd\\xec\\x64\\xba\\x39\\xe5\\x48\\x25\\x0e\\xae\\xa5\\x51\\x17\\x2e\\x53\\x5c"),
    ...                 ECPoint(b"\\x03l\\x841\\xccx\\xb31w\\xa6\\x0bK\\xcc\\x02\\xba\\xf6\\r\\x05\\xfe\\xe5\\x03\\x8es9\\xd3\\xa6\\x88\\xe3\\x94\\xc2\\xcb\\xd8C")],
    ...                [b'wrongsignature1', b'wrongsignature2'])
    False

    :param pubkeys: a list of public keys
    :type pubkeys: List[ECPoint]
    :param signatures: a list of signatures
    :type signatures: List[bytes]
    :return: a boolean value that represents whether the signatures were validated
    :rtype: bool
    """
    pass


def verify_with_ecdsa(message: bytes, pubkey: ECPoint, signature: bytes, curve: NamedCurve) -> bool:
    """
    Using the elliptic curve, it checks if the signature of the message was originally produced by the public key.

    >>> verify_with_ecdsa(b'unit test', ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'),
    ...                   b'wrong_signature', NamedCurve.SECP256R1)
    False

    :param message: the encrypted message
    :type message: bytes
    :param pubkey: the public key that might have created the item
    :type pubkey: ECPoint
    :param signature: the signature of the item
    :type signature: bytes
    :param curve: the curve that will be used by the ecdsa
    :type curve: NamedCurve
    :return: a boolean value that represents whether the signature is valid
    :rtype: bool
    """
    pass


def murmur32(data: bytes, seed: int) -> bytes:
    """
    Computes the hash value for the specified byte array using the murmur32 algorithm.

    >>> murmur32(b'unit test', 0)
    b"\\x90D'G"

    :param data: the input to compute the hash code for
    :type data: bytes
    :param seed: the seed of the murmur32 hash function
    :type seed: int
    :return: the hash value
    :rtype: bytes
    """
    pass
